import asyncio
import re
from typing import TYPE_CHECKING, List, Optional, Set, Tuple, Type

from lonelypss.util.sync_io import SyncReadableBytesIO
from lonelypss.ws.state import (
    AsyncioWSReceiver,
    InternalLargeMessage,
    InternalMessage,
    InternalMessageType,
    InternalMissedMessage,
    InternalSmallMessage,
)


class SimpleReceiver:
    def __init__(self) -> None:
        self.exact_subscriptions: Set[bytes] = set()
        self.glob_subscriptions: List[Tuple[re.Pattern, str]] = []
        self.receiver_id: Optional[int] = None

        self.queue: asyncio.Queue[InternalMessage] = asyncio.Queue()

    def is_relevant(self, topic: bytes) -> bool:
        if topic in self.exact_subscriptions:
            return True

        try:
            topic_str = topic.decode("utf-8", errors="strict")
        except UnicodeDecodeError:
            return False

        return any(pattern.match(topic_str) for pattern, _ in self.glob_subscriptions)

    async def on_large_exclusive_incoming(
        self,
        stream: SyncReadableBytesIO,
        /,
        *,
        topic: bytes,
        sha512: bytes,
        length: int,
    ) -> int:
        finished = asyncio.Event()
        await self.queue.put(
            InternalLargeMessage(
                type=InternalMessageType.LARGE,
                stream=stream,
                length=length,
                finished=finished,
                topic=topic,
                sha512=sha512,
            )
        )
        await finished.wait()
        return 1

    async def on_small_incoming(
        self,
        data: bytes,
        /,
        *,
        topic: bytes,
        sha512: bytes,
    ) -> int:
        await self.queue.put(
            InternalSmallMessage(
                type=InternalMessageType.SMALL, topic=topic, data=data, sha512=sha512
            )
        )
        return 1

    async def on_missed(self, /, *, topic: bytes) -> int:
        await self.queue.put(
            InternalMissedMessage(type=InternalMessageType.MISSED, topic=topic)
        )
        return 1


if TYPE_CHECKING:
    _: Type[AsyncioWSReceiver] = SimpleReceiver



import base64
import hashlib
import json
import tempfile
import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request, Response

from lonelypss.middleware.config import get_config_from_request
from lonelypss.middleware.ws_receiver import get_ws_receiver_from_request

router = APIRouter()


@router.post(
    "/v1/receive_for_websockets",
)
async def receive_for_websockets(
    request: Request,
    authorization: Annotated[Optional[str], Header()] = None,
    repr_digest: Annotated[Optional[str], Header()] = None,
    x_topic: Annotated[Optional[str], Header()] = None,
) -> Response:
    """As a broadcaster, in order to handle websocket connections, we need to be notified
    about messages that were sent to other broadcasters. To facilitate this, the broadcaster
    acts as a subscriber for itself, using this endpoint to receive messages, then forwards
    these to an in-memory structure to fan it out to all the websocket connections.

    This is not the endpoint that subscribers use to notify broadcasters. Use `/v1/notify` for
    that.
    """
    config = get_config_from_request(request)
    receiver = get_ws_receiver_from_request(request)

    if repr_digest is None:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "missing repr-digest header"}',
        )

    if x_topic is None:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "missing x-topic header"}',
        )

    try:
        topic = base64.b64decode(x_topic + "==")
    except BaseException:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "invalid x-topic header"}',
        )

    if not receiver.is_relevant(topic):
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "topic not relevant"}',
        )

    expected_digest_b64: Optional[str] = None
    for digest_pair in repr_digest.split(","):
        split_digest_pair = digest_pair.split("=", 1)
        if len(split_digest_pair) != 2:
            continue
        digest_type, digest_value = split_digest_pair
        if digest_type != "sha-512":
            continue

        expected_digest_b64 = digest_value

    if expected_digest_b64 is None:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "missing sha-512 repr-digest"}',
        )

    try:
        expected_digest = base64.b64decode(expected_digest_b64 + "==")
    except BaseException:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "unparseable sha-512 repr-digest (not base64)"}',
        )

    request_url = str(request.url)
    if request_url != receiver.receiver_url:
        return Response(
            status_code=400,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": "invalid receiver URL"}',
        )

    auth_result = await config.is_receive_allowed(
        url=str(request.url),
        topic=topic,
        message_sha512=expected_digest,
        now=time.time(),
        authorization=authorization,
    )
    if auth_result == "unavailable":
        return Response(status_code=503)
    if auth_result != "ok":
        return Response(
            status_code=403,
            headers={"Content-Type": "application/json; charset=utf-8"},
            content=b'{"unsubscribe": true, "reason": '
            + json.dumps(auth_result).encode("utf-8")
            + b"}",
        )

    with tempfile.SpooledTemporaryFile(
        max_size=config.message_body_spool_size, mode="w+b"
    ) as spooled_request_body:
        read_length = 0
        hasher = hashlib.sha512()
        stream_iter = request.stream().__aiter__()
        while True:
            try:
                chunk = await stream_iter.__anext__()
            except StopAsyncIteration:
                break
            hasher.update(chunk)
            read_length += len(chunk)
            spooled_request_body.write(chunk)

        real_digest = hasher.digest()
        if real_digest != expected_digest:
            return Response(
                status_code=403,
                headers={"Content-Type": "application/json; charset=utf-8"},
                content=b'{"unsubscribe": true, "reason": "incorrect sha-512 repr-digest"}',
            )

        spooled_request_body.seek(0)
        if read_length < config.message_body_spool_size:
            small_body = spooled_request_body.read()
            count = await receiver.on_small_incoming(
                small_body, topic=topic, sha512=real_digest
            )
        else:
            count = await receiver.on_large_exclusive_incoming(
                spooled_request_body,
                topic=topic,
                sha512=real_digest,
                length=read_length,
            )
    return Response(
        status_code=200,
        headers={"Content-Type": "application/json; charset=utf-8"},
        content=b'{"subscribers": ' + str(count).encode("ascii") + b"}",
    )



from fastapi.requests import HTTPConnection
from starlette.types import ASGIApp, Receive, Scope, Send

from lonelypss.config.config import Config


class ConfigMiddleware:
    """Injects `lonelypss_config` into the scope of the request. Required for all
    the routes to function correctly.
    """

    def __init__(self, app: ASGIApp, config: Config):
        self.app = app
        self.config = config

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["lonelypss_config"] = self.config
        await self.app(scope, receive, send)


def get_config_from_request(request: HTTPConnection) -> Config:
    """Retrieves the `lonelypss_config` from the request scope. This is only
    available if the `ConfigMiddleware` is being used
    """
    try:
        return request.scope["lonelypss_config"]
    except KeyError:
        raise Exception("ConfigMiddleware not injecting into request scope")



from typing import TYPE_CHECKING

from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.state import SimplePendingSendPreFormatted, StateOpen


async def send_simple_pending_send_pre_formatted(
    state: StateOpen, message: SimplePendingSendPreFormatted
) -> None:
    """Sends the message to the ASGI server for forwarding"""
    await state.websocket.send_bytes(message.data)


if TYPE_CHECKING:
    _: Sender[SimplePendingSendPreFormatted] = send_simple_pending_send_pre_formatted



import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.missed import B2S_Missed, serialize_b2s_missed

from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypss.ws.state import (
    InternalMissedMessage,
    StateOpen,
)


async def send_internal_missed_message(
    state: StateOpen, message: InternalMissedMessage
) -> None:
    """Notifies the subscriber that the broadcaster may have failed to
    send them a message on a topic they are subscribed to. The standard
    reason this is sent is if there are multiple broadcasters Alice and Bob,
    the subscriber is connected to Alice, and Bob received a message but
    could not tell Alice about it.
    """
    authorization = await state.broadcaster_config.authorize_missed(
        recovery=make_for_send_websocket_url_and_change_counter(state),
        topic=message.topic,
        now=time.time(),
    )
    await state.websocket.send_bytes(
        serialize_b2s_missed(
            B2S_Missed(
                type=BroadcasterToSubscriberStatefulMessageType.MISSED,
                authorization=authorization,
                topic=message.topic,
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        )
    )


if TYPE_CHECKING:
    _: Sender[InternalMissedMessage] = send_internal_missed_message



from lonelypss.config.config import Config


async def setup_config(config: Config) -> None:
    """Convenience function to setup the configuration (similiar idea to aenter)"""
    await config.setup_to_broadcaster_auth()
    try:
        await config.setup_to_subscriber_auth()
        try:
            await config.setup_db()
        except BaseException:
            await config.teardown_to_subscriber_auth()
            raise
    except BaseException:
        await config.teardown_to_broadcaster_auth()
        raise


async def teardown_config(config: Config) -> None:
    """Convenience function to teardown the configuration (similiar idea to aenter)"""
    try:
        await config.teardown_db()
    finally:
        try:
            await config.teardown_to_subscriber_auth()
        finally:
            await config.teardown_to_broadcaster_auth()



# ruff: noqa: T201
import json
import os
import secrets
from argparse import ArgumentParser
from typing import Literal, Optional, Set


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument(
        "--setup",
        action="store_true",
        help="Always required for clarity of the operation.",
    )
    parser.add_argument(
        "--db",
        default="sqlite",
        choices=["sqlite", "rqlite"],
        help="Which backing database to use",
    )
    parser.add_argument(
        "--to-broadcasters-auth",
        default="hmac",
        choices=["hmac", "token", "none"],
        help="How to verify requests to the broadcaster",
    )
    parser.add_argument(
        "--to-broadcasters-secret",
        help="If specified, the secret to use for requests to broadcasters. Ignored unless the to-broadcaster auth strategy requires a secret (hmac, token)",
    )
    parser.add_argument(
        "--to-subscribers-auth",
        default="hmac",
        choices=["hmac", "token", "none"],
        help="How to verify requests to subscribers",
    )
    parser.add_argument(
        "--to-subscribers-secret",
        help="If specified, the secret to use for requests to subscribers. Ignored unless the to-subscriber auth strategy requires a secret (hmac, token)",
    )
    args = parser.parse_args()
    if not args.setup:
        raise Exception("must provide --setup")

    setup_locally(
        db=args.db,
        to_broadcasters_auth=args.to_broadcasters_auth,
        to_broadcasters_secret=args.to_broadcasters_secret,
        to_subscribers_auth=args.to_subscribers_auth,
        to_subscribers_secret=args.to_subscribers_secret,
    )


def setup_locally(
    *,
    db: Literal["sqlite", "rqlite"],
    to_broadcasters_auth: Literal["hmac", "token", "none"],
    to_broadcasters_secret: Optional[str],
    to_subscribers_auth: Literal["hmac", "token", "none"],
    to_subscribers_secret: Optional[str],
) -> None:
    print(
        "httppubserver - Setup\n"
        f"  - db: {db}\n"
        f"  - to-broadcasters-auth: {to_broadcasters_auth}\n"
        f"  - to-broadcasters-secret: {'not specified' if to_broadcasters_secret is None else 'specified'}\n"
        f"  - to-subscribers-auth: {to_subscribers_auth}\n"
        f"  - to-subscribers-secret: {'not specified' if to_subscribers_secret is None else 'specified'}"
    )

    print("Prechecking...")
    for file in [
        "broadcaster-secrets.json",
        "subscriber-secrets.json",
        "main.py",
        "requirements.txt",
    ]:
        if os.path.exists(file):
            raise Exception(f"{file} already exists, refusing to overwrite")

    print("Storing secrets...")
    if to_broadcasters_secret is None:
        to_broadcasters_secret = secrets.token_urlsafe(64)

    if to_subscribers_secret is None:
        to_subscribers_secret = secrets.token_urlsafe(64)

    auth_for_requests_to_broadcasters = (
        {
            "type": to_broadcasters_auth,
            "secret": to_broadcasters_secret,
        }
        if to_broadcasters_auth != "none"
        else None
    )
    auth_for_requests_to_subscribers = (
        {
            "type": to_subscribers_auth,
            "secret": to_subscribers_secret,
        }
        if to_subscribers_auth != "none"
        else None
    )

    for target in ["broadcaster", "subscriber"]:
        with open(f"{target}-secrets.json", "w") as f:
            f.write(
                json.dumps(
                    {
                        "version": "2",
                        **(
                            {"to-broadcaster": auth_for_requests_to_broadcasters}
                            if auth_for_requests_to_broadcasters is not None
                            else {}
                        ),
                        **(
                            {"to-subscriber": (auth_for_requests_to_subscribers)}
                            if auth_for_requests_to_subscribers is not None
                            else {}
                        ),
                    },
                    indent=2,
                )
                + "\n"
            )

    print("Building entrypoint...")

    requirements: Set[str] = set()

    if db == "sqlite":
        db_code = 'SqliteDBConfig("subscriptions.db")'
    else:
        db_code = "TODO()"

    auth_setup_code = ""

    need_secrets = False
    if to_broadcasters_auth == "token":
        to_broadcaster_auth_code = (
            "ToBroadcasterTokenAuth(\n"
            '        token=auth_secrets["to-broadcaster"]["secret"],\n'
            "    )"
        )
        need_secrets = True
    elif to_broadcasters_auth == "hmac":
        if db == "sqlite":
            auth_setup_code = (
                "hmac_db: incoming_auth_config.IncomingHmacAuthDBConfig = (\n"
                '        incoming_auth_config.IncomingHmacAuthSqliteDBConfig("recent-hmacs.db")\n'
                "    )"
            )
        to_broadcaster_auth_code = (
            "ToBroadcasterHmacAuth(\n"
            '        secret=auth_secrets["to-broadcaster"]["secret"],\n'
            "        db_config=hmac_db,\n"
            "    )"
        )
        need_secrets = True
    elif to_broadcasters_auth == "none":
        to_broadcaster_auth_code = "ToBroadcasterNoneAuth()"
    else:
        to_broadcaster_auth_code = "TODO()"

    if to_subscribers_auth == "token":
        to_subscriber_auth_code = 'ToSubscriberTokenAuth(\n        auth_secrets["to-subscriber"]["secret"]\n    )'
        need_secrets = True
    elif to_subscribers_auth == "hmac":
        if db == "sqlite":
            if to_broadcasters_auth != "hmac":
                auth_setup_code = (
                    "hmac_db = incoming_auth_config.IncomingHmacAuthSqliteDBConfig(\n"
                    '        "recent-hmacs.db"\n'
                    "    )"
                )
            else:
                auth_setup_code += "\n    hmac_db = incoming_auth_config.IncomingHmacAuthDBReentrantConfig(hmac_db)"

        to_subscriber_auth_code = (
            "ToSubscriberHmacAuth(\n"
            '        secret=auth_secrets["to-subscriber"]["secret"],\n'
            "        db_config=hmac_db,\n"
            "    )"
        )
    elif to_subscribers_auth == "none":
        to_subscriber_auth_code = "ToSubscriberNoneAuth()"
    else:
        to_subscriber_auth_code = "TODO()"

    load_auth_secrets = (
        ""
        if not need_secrets
        else """
    with open("broadcaster-secrets.json", "r") as f:
        auth_secrets = json.load(f)
"""
    )
    import_json = "import json\n" if need_secrets else ""

    import_config = "\n".join(
        sorted(
            [
                f"import lonelypsp.auth.helpers.{to_broadcasters_auth}_auth_config as incoming_auth_config",
                f"import lonelypsp.auth.helpers.{to_subscribers_auth}_auth_config as outgoing_auth_config",
            ]
        )
    )

    if auth_setup_code:
        auth_setup_code = f"\n    {auth_setup_code}"

    with open("main.py", "w") as f:
        f.write(
            f"""{import_json}from contextlib import asynccontextmanager
from typing import AsyncIterator

{import_config}
import lonelypss.config.helpers.{db}_db_config as db_config
from fastapi import FastAPI
from lonelypsp.auth.config import AuthConfigFromParts
from lonelypss.bknd.sweep_missed import sweep_missed
from lonelypss.config.config import (
    CompressionConfigFromParts,
    Config,
    ConfigFromParts,
    GenericConfigFromValues,
    MissedRetryStandard,
)
from lonelypss.config.lifespan import setup_config, teardown_config
from lonelypss.middleware.config import ConfigMiddleware
from lonelypss.middleware.ws_receiver import WSReceiverMiddleware
from lonelypss.router import router as HttpPubSubRouter
from lonelypss.util.ws_receiver import SimpleFanoutWSReceiver


def _make_config() -> Config:{load_auth_secrets}
    db = db_config.{db_code}{auth_setup_code}
    to_broadcaster_auth = incoming_auth_config.{to_broadcaster_auth_code}
    to_subscriber_auth = outgoing_auth_config.{to_subscriber_auth_code}

    return ConfigFromParts(
        auth=AuthConfigFromParts(
            to_broadcaster=to_broadcaster_auth, to_subscriber=to_subscriber_auth
        ),
        db=db,
        generic=GenericConfigFromValues(
            message_body_spool_size=1024 * 1024 * 10,
            outgoing_http_timeout_total=30,
            outgoing_http_timeout_connect=None,
            outgoing_http_timeout_sock_read=5,
            outgoing_http_timeout_sock_connect=5,
            websocket_accept_timeout=2,
            websocket_max_pending_sends=255,
            websocket_max_unprocessed_receives=255,
            websocket_large_direct_send_timeout=0.3,
            websocket_send_max_unacknowledged=3,
            websocket_minimal_headers=True,
            sweep_missed_interval=10,
        ),
        missed=MissedRetryStandard(
            expo_factor=1,
            expo_base=2,
            expo_max=10,
            max_retries=20,
            constant=1,
            jitter=2,
        ),
        compression=CompressionConfigFromParts(
            compression_allowed=True,
            compression_dictionary_by_id=dict(),
            outgoing_max_ws_message_size=16 * 1024 * 1024,
            allow_training=True,
            compression_min_size=32,
            compression_trained_max_size=16 * 1024,
            compression_training_low_watermark=100 * 1024,
            compression_training_high_watermark=10 * 1024 * 1024,
            compression_retrain_interval_seconds=60 * 60 * 60,
            decompression_max_window_size=8 * 1024 * 1024,
        ),
    )


config = _make_config()
fanout = SimpleFanoutWSReceiver(
    receiver_url="http://127.0.0.1:3003/v1/receive_for_websockets",
    recovery="http://127.0.0.1:3003/v1/missed_for_websockets",
    db=config,
)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[None]:
    await setup_config(config)
    try:
        async with fanout, sweep_missed(config):
            yield
    finally:
        await teardown_config(config)


app = FastAPI(lifespan=lifespan)
app.add_middleware(ConfigMiddleware, config=config)
app.add_middleware(WSReceiverMiddleware, ws_receiver=fanout)
app.include_router(HttpPubSubRouter)
app.router.redirect_slashes = False
"""
        )

    with open("requirements.txt", "w") as f:
        f.write("\n".join(list(sorted(requirements))))

    print("Done! Make sure to install from requirements.txt and pip freeze again!")


if __name__ == "__main__":
    main()



from fastapi import APIRouter

import lonelypss.routes.check_subscriptions
import lonelypss.routes.missed_for_websockets
import lonelypss.routes.notify
import lonelypss.routes.receive_for_websockets
import lonelypss.routes.set_subscriptions
import lonelypss.routes.subscribe_exact
import lonelypss.routes.subscribe_glob
import lonelypss.routes.unsubscribe_exact
import lonelypss.routes.unsubscribe_glob
import lonelypss.routes.websocket_endpoint

router = APIRouter()
router.include_router(lonelypss.routes.check_subscriptions.router)
router.include_router(lonelypss.routes.missed_for_websockets.router)
router.include_router(lonelypss.routes.notify.router)
router.include_router(lonelypss.routes.receive_for_websockets.router)
router.include_router(lonelypss.routes.set_subscriptions.router)
router.include_router(lonelypss.routes.subscribe_exact.router)
router.include_router(lonelypss.routes.subscribe_glob.router)
router.include_router(lonelypss.routes.unsubscribe_exact.router)
router.include_router(lonelypss.routes.unsubscribe_glob.router)
router.include_router(lonelypss.routes.websocket_endpoint.router)



import asyncio
import dataclasses
import io
import os
import re
from dataclasses import dataclass
from tempfile import SpooledTemporaryFile, TemporaryFile
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Type,
    cast,
)

from lonelypsp.compat import fast_dataclass

from lonelypss.config.config import DBConfig
from lonelypss.util.sync_io import (
    SyncIOBaseLikeIO,
    SyncReadableBytesIO,
)
from lonelypss.ws.handlers.open.errors import combine_multiple_exceptions

try:
    from glob import translate as _glob_translate  # type: ignore

    def translate(pat: str) -> str:
        return _glob_translate(pat, recursive=True, include_hidden=True)

except ImportError:
    from fnmatch import translate


class BaseWSReceiver(Protocol):
    """Something capable of processing a received message only after it has
    been verified and is ready for rapid consumption (e.g., in memory or on
    a local disk)
    """

    def is_relevant(self, topic: bytes) -> bool:
        """Allows the caller to skip calls to on_large_exclusive_incoming and
        on_small_incoming if the topic is not relevant to this receiver. If
        async is required for this check, it must be done in the implementation
        of on_large_exclusive_incoming and on_small_incoming instead of here.
        """
        ...

    async def on_large_exclusive_incoming(
        self,
        stream: SyncReadableBytesIO,
        /,
        *,
        topic: bytes,
        sha512: bytes,
        length: int,
    ) -> int:
        """Handles the incoming message on the given topic. It can be assumed
        the stream can be consumed quickly and is not being consumed
        concurrently, but it must not be closed, as the caller may afterward
        seek back to the beginning and reuse the stream.

        The implementation should try to return quickly in all circumstances. For
        example, if the goal is to write the data to a websocket, there should be
        a timeout on a send after which point you copy whatever is remaining to
        a place you control then return back so the stream can be used by the next
        receiver (if any)

        Returns the number of subscribers that were forwarded the message
        """

    async def on_small_incoming(
        self,
        data: bytes,
        /,
        *,
        topic: bytes,
        sha512: bytes,
    ) -> int:
        """Handles the incoming message on the given topic. This is used if
        the data is small enough that spooling is not necessary which allows
        us to concurrently call this function safely (since all the arguments
        are immutable)

        Returns the number of subscribers that were forwarded the message
        """

    async def on_missed(
        self,
        /,
        *,
        topic: bytes,
    ) -> int:
        """Handles that this receiver may have missed some messages on the
        given topic

        Returns the number of subscribers that were forwarded the message
        """


class FanoutWSReceiver(BaseWSReceiver, Protocol):
    """Deduplicates messages and forwards onto the receivers. Note that because
    of this deduplication step, the notifications received over the websocket
    connection are different than those received over the http connection when
    using overlapping globs. This is because the deduplication overhead cannot
    be avoided within the broadcaster in order to avoid receiving multiple
    notifications when multiple websockets are using overlapping globs, which
    is much more reasonable/likely than a single subscriber using overlapping
    globs. It would then take even more effort to reduplicate these messages
    to replicate the one-subscriber scenario.

    A simple async-safe implementation is provided via SimpleFanoutWSReceiver
    """

    @property
    def receiver_url(self) -> str:
        """The URL that this receiver expects to receive messages from"""

    @property
    def missed_url(self) -> Optional[str]:
        """The URL that this receiver expects to receive missed messages from"""

    async def register_receiver(self, receiver: BaseWSReceiver) -> int:
        """Registers a receiver to receive messages from this fanout receiver.
        Returns the id that can be passed to `unregister_receiver`
        """

    async def unregister_receiver(self, receiver_id: int) -> None:
        """Unregisters a receiver from receiving messages from this fanout receiver"""

    async def increment_exact(self, topic: bytes, /) -> None:
        """Increments the count of internal for the given exact topic. If this
        causes the first subscriber to be added, must register us as a local subscriber
        """

    async def decrement_exact(self, topic: bytes, /) -> None:
        """Decrements the count of internal for the given exact topic. If this
        causes the last subscriber to be removed, must unregister us as a local subscriber
        """

    async def increment_glob(self, glob: str, /) -> None:
        """Increments the count of internal for the given glob topic. If this
        causes the first subscriber to be added, must register us as a local subscriber
        """

    async def decrement_glob(self, glob: str, /) -> None:
        """Decrements the count of internal for the given glob topic. If this
        causes the last subscriber to be removed, must unregister us as a local subscriber
        """


@dataclass
class ReceiverNode:
    """the linked-list node for SimpleFanoutWSReceiver"""

    previous: Optional["ReceiverNode"]
    """the previous node or None if this is the head of the list"""
    next: Optional["ReceiverNode"]
    """the next node or None if this is the tail of the list"""
    receiver: BaseWSReceiver
    """the receiver"""
    identifier: int
    """the arbitrary identifier assigned by the fanout receiver"""


@fast_dataclass
class GlobSubscription:
    glob: str
    """the original glob pattern"""
    pattern: re.Pattern
    """the regex that can be used to test the glob pattern"""
    count: int
    """the number of subscribers to this glob pattern"""


class SimpleFanoutWSReceiver:
    """Acts as an async context manager, generally should be initialized as
    a global object but then entered within the appropriate lifespan function.

    Allows subscribing before entering, if desired, and will combine errors
    in that case
    """

    def __init__(
        self, receiver_url: str, recovery: Optional[str], db: DBConfig
    ) -> None:
        self.receiver_url = receiver_url
        """the url we can and expect to receive messages on, i.e, how the
        other broadcasters can reach this broadcaster. Typically,
        `http://<broadcaster_ip>:<broadcaster_port>/v1/receive_for_websockets`
        """
        self.recovery = recovery
        """the url we can and expect to receive missed messages on, i.e, how the
        other broadcasters tell this broadcaster they previously failed to reach
        them; None to indicate we are not requesting MISSED messages
        """
        self.receivers_head: Optional[ReceiverNode] = None
        """the receivers are stored as a double-linked list with a dict
        lookup by id; this points to the head of the list
        """
        self.receivers_tail: Optional[ReceiverNode] = None
        """the receivers are stored as a double-linked list with a dict
        lookup by id; this points to the tail of the list
        """
        self.receivers_list: Optional[List[BaseWSReceiver]] = None
        """if a list of just the base receivers has been materialized to speedup
        iteration, that list, otherwise None.
        """
        self.receivers_iter_counter: int = 0
        """how many times the receivers list has been iterated over since it was
        last mutated; after a threshold which is based on the length of the list,
        i.e., the cost to materialize, materializes the list to speedup iteration
        """
        self.receivers_id_counter: int = 0
        """the counter for assigning unique identifiers to receivers"""
        self.receivers_by_id: Dict[int, ReceiverNode] = dict()
        """a lookup from the identifier assigned to a receiver to the corresponding reciever"""
        self.exact_subscriptions: Dict[bytes, int] = dict()
        """the exact subscriptions that have been registered, mapped to the current
        count, which when it hits 0 means the subscription can be removed
        """
        self.glob_subscriptions: Dict[str, GlobSubscription] = dict()
        """the glob subscriptions that have been registered, mapped to the corresponding
        information about the subscription, including the count
        """
        self.db = db
        """the database connection for telling other broadcasters to send messages to us;
        assumed to be setup by something else whenever this object is entered. Assumed
        to be async-safe so this does not need to coordinate with other users
        """
        self.entered: bool = False
        """True if entered, false if not"""
        self._lock: asyncio.Lock = asyncio.Lock()
        """the lock guarding mutations to receivers, exact_subscriptions, and glob_subscriptions"""

    @property
    def missed_url(self) -> Optional[str]:
        return self.recovery

    async def __aenter__(self) -> "SimpleFanoutWSReceiver":
        async with self._lock:
            assert not self.entered, "already entered and not reentrant safe"

            try_unsubscribes: List[bytes] = []
            try_globs: List[str] = []

            try:
                for topic in self.exact_subscriptions:
                    try_unsubscribes.append(topic)
                    result = await self.db.subscribe_exact(
                        url=self.receiver_url, recovery=self.recovery, exact=topic
                    )
                    if result != "success" and result != "conflict":
                        raise ValueError(f"failed to subscribe to {topic!r}: {result}")

                for glob in self.glob_subscriptions.keys():
                    try_globs.append(glob)
                    result = await self.db.subscribe_glob(
                        url=self.receiver_url, recovery=self.recovery, glob=glob
                    )
                    if result != "success" and result != "conflict":
                        raise ValueError(f"failed to subscribe to {glob}: {result}")

            except BaseException as exc:
                exceptions: List[BaseException] = [exc]

                for topic in try_unsubscribes:
                    try:
                        unsub_res = await self.db.unsubscribe_exact(
                            url=self.receiver_url, exact=topic
                        )
                        if unsub_res != "success" and unsub_res != "not_found":
                            raise ValueError(
                                f"failed to unsubscribe from {topic!r}: {result}"
                            )
                    except BaseException as exc:
                        exceptions.append(exc)

                for glob in try_globs:
                    try:
                        unsub_res = await self.db.unsubscribe_glob(
                            url=self.receiver_url, glob=glob
                        )
                        if unsub_res != "success" and unsub_res != "not_found":
                            raise ValueError(
                                f"failed to unsubscribe from {glob}: {result}"
                            )
                    except BaseException as exc:
                        exceptions.append(exc)

                raise combine_multiple_exceptions(
                    "failed to setup subscriptions", exceptions
                )

            self.entered = True
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        async with self._lock:
            assert self.entered, "not entered and not reentrant safe"

            try_unsubscribes: List[bytes] = list(self.exact_subscriptions)
            try_globs: List[str] = list(glob for glob in self.glob_subscriptions.keys())

            self.exact_subscriptions.clear()
            self.glob_subscriptions.clear()

            exceptions: List[BaseException] = []

            for topic in try_unsubscribes:
                try:
                    result = await self.db.unsubscribe_exact(
                        url=self.receiver_url, exact=topic
                    )
                    if result != "success" and result != "not_found":
                        raise ValueError(
                            f"failed to unsubscribe from {topic!r}: {result}"
                        )
                except BaseException as exc:
                    exceptions.append(exc)

            for glob in try_globs:
                try:
                    result = await self.db.unsubscribe_glob(
                        url=self.receiver_url, glob=glob
                    )
                    if result != "success" and result != "not_found":
                        raise ValueError(f"failed to unsubscribe from {glob}: {result}")
                except BaseException as exc:
                    exceptions.append(exc)

            self.entered = False
            if exceptions:
                raise combine_multiple_exceptions(
                    "failed to teardown subscriptions", exceptions
                )

    async def register_receiver(self, receiver: BaseWSReceiver) -> int:
        async with self._lock:
            identifier = self.receivers_id_counter
            self.receivers_id_counter += 1
            self.receivers_list = None
            self.receivers_iter_counter = 0

            node = ReceiverNode(
                previous=self.receivers_tail,
                next=None,
                receiver=receiver,
                identifier=identifier,
            )
            self.receivers_by_id[identifier] = node

            if self.receivers_head is None:
                assert self.receivers_tail is None
                self.receivers_head = node
                self.receivers_tail = node
                return identifier

            assert self.receivers_tail is not None
            self.receivers_tail.next = node
            self.receivers_tail = node
            return identifier

    async def unregister_receiver(self, receiver_id: int) -> None:
        async with self._lock:
            node = self.receivers_by_id.pop(receiver_id)
            self.receivers_list = None
            self.receivers_iter_counter = 0

            # case 1: node is the only node
            if node.previous is None and node.next is None:
                assert self.receivers_head is node
                assert self.receivers_tail is node
                self.receivers_head = None
                self.receivers_tail = None
                return

            # case 2: node is the head, multiple nodes
            if node.previous is None:
                assert node.next is not None
                assert self.receivers_head is node
                self.receivers_head = node.next
                self.receivers_head.previous = None
                return

            # case 3: node is the tail, multiple nodes
            if node.next is None:
                assert node.previous is not None
                assert self.receivers_tail is node
                self.receivers_tail = node.previous
                self.receivers_tail.next = None
                return

            # case 4: node is in the middle, multiple nodes
            node.previous.next = node.next
            node.next.previous = node.previous
            return

    def iter_receivers(self) -> Iterable[BaseWSReceiver]:
        """Returns an iterator over the receivers; this will use a list
        if its available or if enough iterations have been started for it
        to be worth it, otherwise it will iterate over the linked-list
        """
        if self.receivers_list is not None:
            return self.receivers_list
        slow_iterable = (v.receiver for v in self.receivers_by_id.values())
        self.receivers_iter_counter += 1
        if self.receivers_iter_counter > len(self.receivers_by_id):
            self.receivers_list = list(slow_iterable)
            return self.receivers_list
        return slow_iterable

    async def increment_exact(self, topic: bytes, /) -> None:
        async with self._lock:
            old_count = self.exact_subscriptions.get(topic, 0)
            if old_count > 0:
                self.exact_subscriptions[topic] = old_count + 1
                return

            self.exact_subscriptions[topic] = 1
            if not self.entered:
                return

            try:
                result = await self.db.subscribe_exact(
                    url=self.receiver_url, recovery=self.recovery, exact=topic
                )
                if result != "success" and result != "conflict":
                    raise ValueError(f"failed to subscribe to {topic!r}: {result}")
            except BaseException:
                unsub_result = await self.db.unsubscribe_exact(
                    url=self.receiver_url, exact=topic
                )
                if unsub_result != "success" and unsub_result != "not_found":
                    raise ValueError(
                        f"failed to unsubscribe from {topic!r}: {unsub_result}"
                    )
                del self.exact_subscriptions[topic]
                raise

    async def decrement_exact(self, topic: bytes, /) -> None:
        async with self._lock:
            old_count = self.exact_subscriptions.get(topic, 0)
            if old_count == 0:
                return

            if old_count > 1:
                self.exact_subscriptions[topic] = old_count - 1
                return

            del self.exact_subscriptions[topic]
            if not self.entered:
                return

            result = await self.db.unsubscribe_exact(url=self.receiver_url, exact=topic)
            if result != "success" and result != "not_found":
                raise ValueError(f"failed to unsubscribe from {topic!r}: {result}")

    async def increment_glob(self, glob: str, /) -> None:
        async with self._lock:
            subscription = self.glob_subscriptions.get(glob)
            if subscription is not None:
                self.glob_subscriptions[glob] = dataclasses.replace(
                    subscription, count=subscription.count + 1
                )
                return

            pattern = re.compile(translate(glob))
            subscription = GlobSubscription(glob=glob, pattern=pattern, count=1)
            self.glob_subscriptions[glob] = subscription
            if not self.entered:
                return

            try:
                result = await self.db.subscribe_glob(
                    url=self.receiver_url, recovery=self.recovery, glob=glob
                )
                if result != "success" and result != "conflict":
                    raise ValueError(f"failed to subscribe to {glob!r}: {result}")
            except BaseException:
                unsub_result = await self.db.unsubscribe_glob(
                    url=self.receiver_url, glob=glob
                )
                if unsub_result != "success" and unsub_result != "not_found":
                    raise ValueError(
                        f"failed to unsubscribe from {glob!r}: {unsub_result}"
                    )
                del self.glob_subscriptions[glob]
                raise

    async def decrement_glob(self, glob: str, /) -> None:
        async with self._lock:
            subscription = self.glob_subscriptions.get(glob)
            if subscription is None:
                return

            if subscription.count > 1:
                self.glob_subscriptions[glob] = dataclasses.replace(
                    subscription, count=subscription.count - 1
                )
                return

            del self.glob_subscriptions[glob]
            if not self.entered:
                return

            result = await self.db.unsubscribe_glob(url=self.receiver_url, glob=glob)
            if result != "success" and result != "not_found":
                raise ValueError(f"failed to unsubscribe from {glob!r}: {result}")

    def is_relevant(self, topic: bytes) -> bool:
        if topic in self.exact_subscriptions:
            return True

        try:
            topic_str = topic.decode("utf-8", errors="strict")
        except UnicodeDecodeError:
            return False

        return any(
            glob.pattern.match(topic_str) for glob in self.glob_subscriptions.values()
        )

    async def on_large_exclusive_incoming(
        self,
        stream: SyncReadableBytesIO,
        /,
        *,
        topic: bytes,
        sha512: bytes,
        length: int,
    ) -> int:
        is_tellable: bool
        is_seekable: bool

        if isinstance(stream, (io.IOBase, SpooledTemporaryFile)):
            is_tellable = True
            is_seekable = True
        else:
            has_tell = hasattr(stream, "tell") and callable(
                cast(SyncIOBaseLikeIO, stream).tell
            )
            has_tellable = hasattr(stream, "tellable") and callable(
                cast(Any, stream).tellable
            )
            is_tellable = has_tell and (
                (not has_tellable) or cast(Any, stream).tellable()
            )

            has_seek = hasattr(stream, "seek") and callable(
                cast(SyncIOBaseLikeIO, stream).seek
            )
            has_seekable = hasattr(stream, "seekable") and callable(
                cast(Any, stream).seekable
            )
            is_seekable = has_seek and (
                (not has_seekable) or cast(Any, stream).seekable()
            )

        need_close: bool
        std_io: SyncIOBaseLikeIO

        if is_tellable and is_seekable:
            need_close = False
            std_io = cast(SyncIOBaseLikeIO, stream)
        else:
            need_close = True
            std_io = TemporaryFile("w+b")
            try:
                while True:
                    data = stream.read(io.DEFAULT_BUFFER_SIZE)
                    if not data:
                        break
                    std_io.write(data)
                std_io.seek(0, os.SEEK_SET)
            except BaseException:
                std_io.close()
                raise

        try:
            start_pos = std_io.tell()
            count = 0
            for receiver in self.iter_receivers():
                if not receiver.is_relevant(topic):
                    continue
                std_io.seek(start_pos, os.SEEK_SET)
                count += await receiver.on_large_exclusive_incoming(
                    stream, topic=topic, sha512=sha512, length=length
                )
            return count
        finally:
            if need_close:
                std_io.close()

    async def on_small_incoming(
        self,
        data: bytes,
        /,
        *,
        topic: bytes,
        sha512: bytes,
    ) -> int:
        count = 0
        for receiver in self.iter_receivers():
            if not receiver.is_relevant(topic):
                continue
            count += await receiver.on_small_incoming(data, topic=topic, sha512=sha512)
        return count

    async def on_missed(
        self,
        /,
        *,
        topic: bytes,
    ) -> int:
        count = 0
        for receiver in self.iter_receivers():
            if not receiver.is_relevant(topic):
                continue
            count += await receiver.on_missed(topic=topic)
        return count


if TYPE_CHECKING:
    _: Type[FanoutWSReceiver] = SimpleFanoutWSReceiver



from typing import TYPE_CHECKING, Any, Dict, Union, cast

from lonelypss.ws.handlers.open.senders.protocol import Sender
from lonelypss.ws.handlers.open.senders.send_internal_large_message import (
    send_internal_large_message,
)
from lonelypss.ws.handlers.open.senders.send_internal_missed_message import (
    send_internal_missed_message,
)
from lonelypss.ws.handlers.open.senders.send_internal_small_message import (
    send_internal_small_message,
)
from lonelypss.ws.handlers.open.senders.send_simple_pending_send_pre_formatted import (
    send_simple_pending_send_pre_formatted,
)
from lonelypss.ws.handlers.open.senders.send_waiting_internal_spooled_large_message import (
    send_waiting_internal_spooled_large_message,
)
from lonelypss.ws.state import (
    InternalLargeMessage,
    InternalMessageType,
    InternalMissedMessage,
    InternalSmallMessage,
    SimplePendingSendPreFormatted,
    SimplePendingSendType,
    StateOpen,
    WaitingInternalMessageType,
    WaitingInternalSpooledLargeMessage,
)

Sendable = Union[
    InternalSmallMessage,
    InternalMissedMessage,
    InternalLargeMessage,
    WaitingInternalSpooledLargeMessage,
    SimplePendingSendPreFormatted,
]

SENDERS: Dict[Any, Any] = {
    InternalMessageType.SMALL: send_internal_small_message,
    InternalMessageType.MISSED: send_internal_missed_message,
    InternalMessageType.LARGE: send_internal_large_message,
    WaitingInternalMessageType.SPOOLED_LARGE: send_waiting_internal_spooled_large_message,
    SimplePendingSendType.PRE_FORMATTED: send_simple_pending_send_pre_formatted,
}


async def send_any(
    state: StateOpen,
    message: Sendable,
) -> None:
    """The target for `send_task` on StateOpen. This will write to the websocket and
    expect that nothing else is doing so.

    When sending an internal large message that is unspooled, this will handle
    setting the finished event as quickly as possible; to facilitate this, it
    will spool the message if it detects its taking too long to push the message
    to the ASGI server.
    """
    await cast(Sender[Sendable], SENDERS[message.type])(state, message)


if TYPE_CHECKING:
    _: Sender[Sendable] = send_any



import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeGlob,
    serialize_b2s_confirm_unsubscribe_glob,
)
from lonelypsp.stateful.messages.unsubscribe import S2B_UnsubscribeGlob

from lonelypss.ws.handlers.open.errors import AuthRejectedException
from lonelypss.ws.handlers.open.processors.protocol import S2B_MessageProcessor
from lonelypss.ws.handlers.open.send_simple_asap import send_simple_asap
from lonelypss.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypss.ws.state import StateOpen


async def process_unsubscribe_glob(
    state: StateOpen, message: S2B_UnsubscribeGlob
) -> None:
    """Processes a request by the subscriber to unsubscribe from utf-8 decodable
    topics which match the given glob pattern, given that they subscribed to this
    exact glob pattern before.
    """
    url = make_for_receive_websocket_url_and_change_counter(state)
    auth_at = time.time()
    auth_result = await state.broadcaster_config.is_subscribe_glob_allowed(
        url=url,
        recovery=None,
        glob=message.glob,
        now=auth_at,
        authorization=message.authorization,
    )
    if auth_result != "ok":
        raise AuthRejectedException(f"subscribe exact: {auth_result}")

    for idx, (_, glob) in enumerate(state.my_receiver.glob_subscriptions):
        if glob == message.glob:
            break
    else:
        raise Exception("not subscribed to glob pattern")

    state.my_receiver.glob_subscriptions.pop(idx)
    await state.internal_receiver.decrement_glob(message.glob)
    send_simple_asap(
        state,
        serialize_b2s_confirm_unsubscribe_glob(
            B2S_ConfirmUnsubscribeGlob(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_GLOB,
                glob=message.glob,
            ),
            minimal_headers=state.broadcaster_config.websocket_minimal_headers,
        ),
    )


if TYPE_CHECKING:
    _: S2B_MessageProcessor[S2B_UnsubscribeGlob] = process_unsubscribe_glob



import time
from typing import Annotated, Optional

from fastapi import APIRouter, Header, Request, Response

from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.async_io import async_read_exact
from lonelypss.util.request_body_io import AsyncIterableAIO

router = APIRouter()


@router.post(
    "/v1/subscribe/glob",
    status_code=202,
    responses={
        "400": {"description": "The body was not formatted correctly"},
        "401": {"description": "Authorization header is required but not provided"},
        "403": {"description": "Authorization header is provided but invalid"},
        "409": {"description": "The subscription already exists"},
        "500": {"description": "Unexpected error occurred"},
        "503": {"description": "Service is unavailable, try again soon"},
    },
)
async def subscribe(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Subscribes the given URL to any message posted to a topic matching the
    indicated glob pattern. The body is formatted as follows:

    - 2 bytes (N): the length of the url, big-endian, unsigned
    - N bytes: the url. must be valid utf-8
    - 2 bytes (M): the length of the glob.
    - M bytes: the glob. must be valid utf-8
    - 2 bytes (R): either 0, to indicate no missed messages are desired, or the length
      of the url to post missed messages to, big-endian, unsigned
    - R bytes: the url to post missed messages to, utf-8 encoded

    NOTE: if you want to use the same path and glob for multiple subscriptions
    to get multiple notifications, you can include a hash that disambiguates them,
    for example http://192.0.2.0:8080/#uid=abc123

    The response has an arbitrary body (generally empty) and one of the
    following status codes:

    - 202 Accepted: the subscription was added
    - 400 Bad Request: the body was not formatted correctly
    - 401 Unauthorized: authorization is required but not provided
    - 403 Forbidden: authorization is provided but invalid
    - 409 Conflict: the subscription already exists
    - 500 Internal Server Error: unexpected error occurred
    - 503 Service Unavailable: servce (generally, database) is unavailable
    """
    config = get_config_from_request(request)

    try:
        stream = request.stream()
        try:
            body = AsyncIterableAIO(stream.__aiter__())

            url_length_bytes = await async_read_exact(body, 2)
            url_length = int.from_bytes(url_length_bytes, "big")
            url_bytes = await async_read_exact(body, url_length)
            url = url_bytes.decode("utf-8")

            glob_length_bytes = await async_read_exact(body, 2)
            glob_length = int.from_bytes(glob_length_bytes, "big")
            glob_bytes = await async_read_exact(body, glob_length)
            glob = glob_bytes.decode("utf-8")

            recovery_url_length_bytes = await async_read_exact(body, 2)
            recovery_url_length = int.from_bytes(recovery_url_length_bytes, "big")
            recovery_url: Optional[str] = None
            if recovery_url_length > 0:
                recovery_url_bytes = await async_read_exact(body, recovery_url_length)
                recovery_url = recovery_url_bytes.decode("utf-8")
        finally:
            await stream.aclose()
    except ValueError:
        return Response(status_code=400)

    auth_at = time.time()
    auth_result = await config.is_subscribe_glob_allowed(
        url=url,
        recovery=recovery_url,
        glob=glob,
        now=auth_at,
        authorization=authorization,
    )

    if auth_result == "unauthorized":
        return Response(status_code=401)
    elif auth_result == "forbidden":
        return Response(status_code=403)
    elif auth_result == "unavailable":
        return Response(status_code=503)
    elif auth_result != "ok":
        return Response(status_code=500)

    db_result = await config.subscribe_glob(url=url, recovery=recovery_url, glob=glob)

    if db_result == "conflict":
        return Response(status_code=409)
    elif db_result == "unavailable":
        return Response(status_code=503)
    elif db_result != "success":
        return Response(status_code=500)

    return Response(status_code=202)



import asyncio

from lonelypss.ws.state import (
    SimplePendingSendPreFormatted,
    SimplePendingSendType,
    StateOpen,
)


def send_simple_asap(state: StateOpen, data: bytes) -> None:
    """Queues a simple message to be sent to the subscriber as soon as possible"""

    if state.send_task is None:
        state.send_task = asyncio.create_task(state.websocket.send_bytes(data))
        return

    state.unsent_messages.append(
        SimplePendingSendPreFormatted(
            type=SimplePendingSendType.PRE_FORMATTED, data=data
        )
    )



import io
import sys
from typing import Any, Union

from lonelypss.util.sync_io import SyncIOBaseLikeIO

if sys.version_info <= (3, 10):
    from tempfile import SpooledTemporaryFile

    ChildType = Union[io.IOBase, SpooledTemporaryFile[bytes], SyncIOBaseLikeIO]
else:
    ChildType = Union[io.IOBase, SyncIOBaseLikeIO]


def _noop(*args: Any, **kwargs: Any) -> None:
    pass


class CloseGuardedIO(io.IOBase):
    """Wraps an IO object, except for close() which does nothing"""

    def __init__(self, child: ChildType) -> None:
        self._child = child

    def __getattribute__(self, name: str) -> Any:
        if name == "_child":
            return super().__getattribute__(name)
        if name == "close":
            return _noop
        return getattr(self._child, name)



import asyncio
import enum
import importlib
import random
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
)

from lonelypsp.auth.config import AuthConfig
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateless.make_strong_etag import StrongEtag

from lonelypss.config.set_subscriptions_info import SetSubscriptionsInfo

try:
    import zstandard
except ImportError:
    ...


class SubscriberInfoType(enum.Enum):
    EXACT = enum.auto()
    GLOB = enum.auto()
    UNAVAILABLE = enum.auto()


@fast_dataclass
class SubscriberInfoExact:
    type: Literal[SubscriberInfoType.EXACT]
    """Indicates we found a subscriber on this exact topic"""
    url: str
    """The url to reach the subscriber"""
    recovery: Optional[str]
    """The recovery url to later tell the subscriber if we fail to
    send them this message, if any
    """


@fast_dataclass
class SubscriberInfoGlob:
    type: Literal[SubscriberInfoType.GLOB]
    """Indicates we found a subscriber for this topic via a matching glob subscription"""
    glob: str
    """The glob that matched the topic"""
    url: str
    """The url to reach the subscriber"""
    recovery: Optional[str]
    """The recovery url to later tell the subscriber if we fail to
    send them this message, if any
    """


@fast_dataclass
class SubscriberInfoUnavailable:
    type: Literal[SubscriberInfoType.UNAVAILABLE]
    """Indicates that the database for subscriptions is unavailable
    and the request should be aborted with a 503
    """


SubscriberInfo = Union[
    SubscriberInfoExact, SubscriberInfoGlob, SubscriberInfoUnavailable
]


@fast_dataclass
class MissedInfo:
    """information about a url we are trying to send a missed message to"""

    topic: bytes
    """the topic that the message was supposed to be sent to"""
    attempts: int
    """how many times we have tried to send this message"""
    next_retry_at: float
    """the next time we should try to send this message"""
    subscriber_info: Union[SubscriberInfoExact, SubscriberInfoGlob]
    """the subscriber that was supposed to receive the message; if deleted,
    the missed row MUST also be deleted
    """


@fast_dataclass
class MutableMissedInfo:
    """just the actually changable parts of MissedInfo after an attempt"""

    attempts: int
    """how many times we have tried to send this message"""
    next_retry_at: float
    """the next time we should try to send this message"""


class LockedMissedInfo(Protocol):
    """Used in `get_overdue_missed_with_lock`; a MissedInfo item that is temporarily
    prevented from being returned in concurrent calls to `get_overdue_missed_with_lock`
    (both in other processes connected to the same database and in other
    coroutines in the same async thread) until the lock time has expired
    or `release` has been called.
    """

    @property
    def info(self) -> MissedInfo:
        """the potentially still protected MissedInfo"""

    async def release(
        self,
        /,
        *,
        new_info: Optional[MutableMissedInfo] = None,
    ) -> Literal["conflict", "unavailable", "ok"]:
        """Releases the lock on this item and sets the new information or
        deletes the record based on `new_info`. This should not raise an
        exception if the lock has already expired or `release` has already
        been called, instead, return `conflict`

        Args:
            new_info (Optional[MutableMissedInfo], None): the new information
                to set, or None to delete the record

        Returns:
            `conflict`: if the record is no longer the same as it was so
                this did nothing
            `unavailable`: if the database for missed messages is unavailable
            `ok`: the record was the same as it was and it was updated/deleted
                as requested
        """


class DBConfig(Protocol):
    async def setup_db(self) -> None:
        """Prepares the database for use. If the database is not re-entrant, it must
        check for re-entrant calls and error out
        """

    async def teardown_db(self) -> None:
        """Cleans up the database after use. This is called when the server is done
        using the database, and should release any resources it acquired during
        `setup_db`.
        """

    async def subscribe_exact(
        self, /, *, url: str, recovery: Optional[str], exact: bytes
    ) -> Literal["success", "conflict", "unavailable"]:
        """Subscribes the given URL to the given exact match.

        Args:
            url (str): the url that will receive notifications
            recovery (str, None): if the subscriber wishes for a MISSED message
                to be posted to a url when the broadcaster fails to send them a
                message on this topic (later, with more spread out retries),
                this is the url that will receive that message. If None, no
                MISSED message will be sent. NOTE: missed messages do not
                contain the actual message missed
            exact (bytes): the exact topic they want to receive messages from

        Returns:
            `success`: if the subscription was added
            `conflict`: if the subscription already exists. the recovery url MUST NOT
                have been changed (if it differs)
            `unavailable`: if a service is required to check this isn't available
        """

    async def unsubscribe_exact(
        self, /, *, url: str, exact: bytes
    ) -> Literal["success", "not_found", "unavailable"]:
        """Unsubscribes the given URL from the given exact match

        Args:
            url (str): the url that will receive notifications
            exact (bytes): the exact topic they want to receive messages from

        Returns:
            `success`: if the subscription was removed
            `not_found`: if the subscription didn't exist
            `unavailable`: if the database for subscriptions is unavailable
        """

    async def subscribe_glob(
        self, /, *, url: str, recovery: Optional[str], glob: str
    ) -> Literal["success", "conflict", "unavailable"]:
        """Subscribes the given URL to the given glob-style match

        Args:
            url (str): the url that will receive notifications
            recovery (str, None): if the subscriber wishes for a MISSED message
                to be posted to a url when the broadcaster fails to send them a
                message on this glob (later, with more spread out retries), this
                is the url that will receive that message. If None, no MISSED
                message will be sent. NOTE: missed messages do not contain the
                actual message missed
            glob (str): a glob for the topics that they want to receive notifications from

        Returns:
            `success`: if the subscription was added
            `conflict`: if the subscription already exists. the recovery url MUST NOT
                have been changed (if it differs)
            `unavailable`: if the database for subscriptions is unavailable
        """

    async def unsubscribe_glob(
        self, /, *, url: str, glob: str
    ) -> Literal["success", "not_found", "unavailable"]:
        """Unsubscribes the given URL from the given glob-style match

        Args:
            url (str): the url that will receive notifications
            glob (str): a glob for the topics that they want to receive notifications from

        Returns:
            `success`: if the subscription was removed
            `not_found`: if the subscription didn't exist
            `unavailable`: if the database for subscriptions is unavailable
        """

    async def check_subscriptions(self, /, *, url: str) -> StrongEtag:
        """Determines what the current subscriptions are for a given URL, computing
        the strong etag corresponding to that url and set of subscriptions, and
        returning that.

        NOTE: the strong etag format is designed to be computable without loading
        all the subscriptions into memory but does require a linear pass over the
        relevant subscriptions.

        NOTE: typically, paginate through results and use create_strong_etag_generator

        NOTE: if the subscriptions are mutated while this is running the result must
        be an etag that includes every item that was in the database the entire
        iteration, must not include any item that was not in the database the entire
        iteration, and must not include duplicates. Items that were added or removed
        during the iteration may be arbitrarily included or excluded from the etag.

        Args:
            url (str): the url that we are checking

        Returns:
            StrongEtag: the strong etag for the subscriptions
        """

    async def set_subscriptions(
        self,
        /,
        *,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
    ) -> Literal["success", "unavailable"]:
        """Sets the subscriptions for a given URL to the given topics and globs.
        This does not need to be completely atomic, though it is desirable. If
        it is not atomic, the rules within lonelypsp's documentation for
        `SET_SUBSCRIPTION` must be followed

        Args:
            url (str): the url that will receive notifications
            strong_etag (StrongEtag): the etag that was returned from the last check_subscriptions call
            subscriptions (SetSubscriptionsInfo): the topics and globs that the
                subscriber wants to receive. this list may be large, so a more
                restrictive interface is provided to provide flexibility in its
                implementation. the list is guarranteed to be in ascending lexicographic
                order for the topics/globs

        Returns:
            `success`: if the subscriptions were set
            `conflict`: if the etag didn't match the current state of the database
            `unavailable`: if the database for subscriptions is unavailable
        """

    def get_subscribers(self, /, *, topic: bytes) -> AsyncIterable[SubscriberInfo]:
        """Streams back the subscriber urls that match the given topic. We will post messages
        to these urls as they are provided. This should return duplicates if multiple subscriptions
        match with the same url.

        Args:
            topic (bytes): the topic that we are looking for

        Yields:
            (SubscriberInfo): the subscriber that was found, or a special value indicating
                that the database is unavailable.
        """

    async def upsert_missed(
        self, /, *, info: MissedInfo
    ) -> Literal["success", "unavailable"]:
        """Updates the given missed information, which should be looked up using
        the subscriber information, with the new attempt/next attempt time. If
        the subscriber is a GLOB subscriber, the topic should also be updated

        Should clear any locks from `get_overdue_missed_with_lock` for the same
        `info.subscriber_info.url` and `info.topic`

        Args:
            info (MissedInfo): the information about the missed message

        Returns:
            `success`: if the missed message was updated
            `unavailable`: if the database for missed messages is unavailable
        """

    def get_overdue_missed_with_lock(
        self, /, *, now: float
    ) -> AsyncIterable[LockedMissedInfo]:
        """Streams back the missed messages that are overdue for retrying. This
        has important concurrency requirements: when this is called in parallel,
        async or from different processes, it should not repeat messages until
        the lock time is elapsed without `__aexit__` being called on the item.
        The choice of lock time is implementation-defined, but should be at least
        enough time to attempt contacting the subscriber and getting a response.

        Args:
            now (float): the current time in seconds since the epoch

        Yields:
            (LockedMissedInfo): the missed message that is overdue
        """


class GenericConfig(Protocol):
    @property
    def message_body_spool_size(self) -> int:
        """If the message body exceeds this size we always switch to a temporary file.

        In general, unless there is another specific configuration option for it, this
        is the maximum size of any single arbitrary length item (e.g., the decompressed
        body of a compressed message) that is held in memory before spooling to file.
        """

    @property
    def outgoing_http_timeout_total(self) -> Optional[float]:
        """The total timeout for outgoing http requests in seconds"""

    @property
    def outgoing_http_timeout_connect(self) -> Optional[float]:
        """The timeout for connecting to the server in seconds, which may include multiple
        socket attempts
        """

    @property
    def outgoing_http_timeout_sock_read(self) -> Optional[float]:
        """The timeout for reading from a socket to the server in seconds before the socket is
        considered dead
        """

    @property
    def outgoing_http_timeout_sock_connect(self) -> Optional[float]:
        """The timeout for a single socket connecting to the server before we give up in seconds"""

    @property
    def websocket_accept_timeout(self) -> Optional[float]:
        """The timeout for accepting a websocket connection in seconds"""

    @property
    def websocket_max_pending_sends(self) -> Optional[int]:
        """The maximum number of pending sends (not yet sent to the ASGI server) before we
        disconnect the websocket forcibly. This mainly protects against (accidental) tarpitting
        when the subscriber cannot keep up
        """

    @property
    def websocket_max_unprocessed_receives(self) -> Optional[int]:
        """The maximum number of unprocessed received websocket messages before we disconnect
        the websocket forcibly. Note that this does not include any websocket messages buffered
        above us (such as in the ASGI server). This is primarily intended to help improve recovery
        when the broadcaster cannot keep up by resetting connections we are very far behind on
        """

    @property
    def websocket_large_direct_send_timeout(self) -> Optional[float]:
        """How long we are willing to wait for a websocket.send to complete while holding
        an exclusive file handle to the message being sent before copying the remainder of
        the message to memory and progressing the other sends. This value is in seconds,
        and a reasonable choice is 0.3 (300ms)

        A value of 0 means it must complete within one event loop
        """

    @property
    def websocket_send_max_unacknowledged(self) -> Optional[int]:
        """The maximum number of NOTIFY/NOTIFY STREAM messages we will have outgoing before
        waiting for them to be acknowledged by the client. This is intended to help alleviate
        issues related to clients that cannot keep up due to processing issues rather than
        network issues, as having full websocket buffers tends to result in a poor debugging
        experience.

        A reasonable value is 3
        """

    @property
    def websocket_minimal_headers(self) -> bool:
        """True if all messages from the broadcaster to the subscriber should use
        minimal headers, which are faster to parse and more compact but require
        that the subscriber and broadcaster precisely agree on the headers for
        each message. False if all messages from the broadcaster to the
        subscriber use expanded headers, which are more flexible and easier to
        debug but slower to parse and more verbose.

        If you are trying to understand the lonelypss protocol via something
        like wireshark, setting this to False will make messages somewhat easier
        to understand.

        Note that broadcasters and subscribers do not need to agree on this
        setting. It is ok if the broadcaster is sending expanded headers and the
        subscriber is sending minimal headers, or vice versa, as this only
        configures the outgoing messages but they both always accept either
        version for incoming messages.

        Generally, this should be True except when in the process of updating
        the lonelypss/lonelypsc libraries, in which case it should be changed to
        false on the broadcaster and subscribers, then they should be updated
        one at a time, then set to true.
        """

    @property
    def sweep_missed_interval(self) -> float:
        """The interval in seconds between sweeps of pending MISSED messages,
        which are retried for a significant period of time (more than enough
        to get past any transient network issues) so that subscribers can
        trigger their reboot recovery mechanism if they did not detect that
        they were not reachable
        """


class GenericConfigFromValues:
    """Convenience class that allows you to create a GenericConfig protocol
    satisfying object from values"""

    def __init__(
        self,
        message_body_spool_size: int,
        outgoing_http_timeout_total: Optional[float],
        outgoing_http_timeout_connect: Optional[float],
        outgoing_http_timeout_sock_read: Optional[float],
        outgoing_http_timeout_sock_connect: Optional[float],
        websocket_accept_timeout: Optional[float],
        websocket_max_pending_sends: Optional[int],
        websocket_max_unprocessed_receives: Optional[int],
        websocket_large_direct_send_timeout: Optional[float],
        websocket_send_max_unacknowledged: Optional[int],
        websocket_minimal_headers: bool,
        sweep_missed_interval: float,
    ):
        self.message_body_spool_size = message_body_spool_size
        self.outgoing_http_timeout_total = outgoing_http_timeout_total
        self.outgoing_http_timeout_connect = outgoing_http_timeout_connect
        self.outgoing_http_timeout_sock_read = outgoing_http_timeout_sock_read
        self.outgoing_http_timeout_sock_connect = outgoing_http_timeout_sock_connect
        self.websocket_accept_timeout = websocket_accept_timeout
        self.websocket_max_pending_sends = websocket_max_pending_sends
        self.websocket_max_unprocessed_receives = websocket_max_unprocessed_receives
        self.websocket_large_direct_send_timeout = websocket_large_direct_send_timeout
        self.websocket_send_max_unacknowledged = websocket_send_max_unacknowledged
        self.websocket_minimal_headers = websocket_minimal_headers
        self.sweep_missed_interval = sweep_missed_interval


class MissedRetryConfig(Protocol):
    async def get_delay_for_next_missed_retry(
        self, /, *, receive_url: str, missed_url: str, topic: bytes, attempts: int
    ) -> Optional[float]:
        """Determines the number of fractional seconds to wait before retrying
        a missed message sent to a subscriber that failed to receive a message
        on a topic.

        Essentially, the flow on failure when a recovery url is specified is

        ```
        Broadcaster -> Subscriber: RECEIVE (error)
        (get_delay_for_next_missed_retry(attempts=0) seconds pass)
        Broadcaster -> Subscriber: MISSED (error)
        (get_delay_for_next_missed_retry(attempts=1) seconds pass)
        Broadcaster -> Subscriber: MISSED (error)
        (get_delay_for_next_missed_retry(attempts=2) seconds pass)
        ...
        ```

        May return None to never retry again.

        Args:
            receive_url (str): the url that was supposed to get the RECEIVE
            missed_url (str): the url that gets the MISSED
            topic (bytes): the topic the message was posted to
            attempts (int): the number of attempts that have been made so far
                to send the MISSED message

        Returns:
            None: no more retries
            float: the minimum time in seconds to wait before retrying
        """


class MissedRetryStandard:
    """Standard implementation of the MissedRetryConfig protocol

    `expo_factor * (expo_base ** min(expo_max, attempts)) + constant + random() * jitter`
    """

    def __init__(
        self,
        expo_factor: float,
        expo_base: float,
        expo_max: int,
        max_retries: int,
        constant: float,
        jitter: float,
    ) -> None:
        """
        `expo_factor * (expo_base ** min(expo_max, attempts)) + constant + random() * jitter`

        Args:
            expo_factor (float): the factor to multiply the base by to get the
                delay between retries
            expo_base (float): the base delay between retries
            expo_max (int): the maximum number of retries before we stop
            max_retries (int): the maximum number of retries before we stop
            constant (float): the constant delay to add to the exponential delay
            jitter (float): the amount of jitter to add to the delay
        """
        self.expo_factor = expo_factor
        self.expo_base = expo_base
        self.expo_max = expo_max
        self.max_retries = max_retries
        self.constant = constant
        self.jitter = jitter

    async def get_delay_for_next_missed_retry(
        self, /, *, receive_url: str, missed_url: str, topic: bytes, attempts: int
    ) -> Optional[float]:
        if attempts >= self.max_retries:
            return None

        return (
            self.expo_factor * (self.expo_base ** min(self.expo_max, attempts))
            + self.constant
            + random.random() * self.jitter
        )


class CompressionConfig(Protocol):
    """Configuration for compression over websockets"""

    @property
    def compression_allowed(self) -> bool:
        """True to allow zstandard compression for websocket messages when supported
        by the client, false to disable service-level compression entirely.
        """

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /
    ) -> "Optional[Tuple[zstandard.ZstdCompressionDict, int]]":
        """If a precomputed zstandard compression dictionary is available with the
        given id, the dictionary and the compression level to use should be
        returned. If the dictionary is not available, return None.

        This is generally only useful if you are using short-lived websocket
        connections where trained dictionaries won't kick in, or you want to go
        through the effort of hand-building a dictionary for a specific
        use-case.

        The returned dict should have its data precomputed as if by `precompute_compress`
        """

    @property
    def outgoing_max_ws_message_size(self) -> Optional[int]:
        """We will try not to send websocket messages over this size. If this is at least
        64kb, then we will guarrantee we won't go over this value.

        Generally, breaking websocket messages apart is redundant: the websocket protocol
        already has a concept of frames which can be used to send messages in parts. Further,
        it's almost certainly more performant to break messages at a lower level in the stack.

        However, in practice, the default settings of most websocket servers and
        clients will not accept arbitrarily large messages, and the entire
        message is often kept in memory, so we break them up to avoid issues. Note that when
        we break messages we will spool to disk as they come in.

        A reasonable value is 16mb

        Return None for no limit
        """

    @property
    def allow_training(self) -> bool:
        """True to allow training dictionaries in websockets, false to completely disable
        that feature
        """

    @property
    def compression_min_size(self) -> int:
        """The smallest message size we will try to compress

        A reasonable size is 32 bytes
        """

    @property
    def compression_trained_max_size(self) -> int:
        """The largest message size we will try to use a custom trained dictionary for; for
        messages larger than this, we will not use a shared compression dictionary to reduce
        overhead, as theres enough context within the message to generate its own dictionary,
        and the relative overhead of including that dictionary will be low.

        A reasonable value is 16kb
        """

    @property
    def compression_training_low_watermark(self) -> int:
        """How much data we get before we make the first pass at the custom compression dictionary

        A reasonable value is 100kb
        """

    async def train_compression_dict_low_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        """Trains a compression dictionary using the samples whose combined size is at least the
        `compression_training_low_watermark` size, then tells us what level compression to use

        Typically, this is something like

        ```python
        import zstandard
        import asyncio

        async def train_compression_dict_low_watermark(samples: List[bytes]) -> zstandard.ZstdCompressionDict:
            zdict = await asyncio.to_thread(
                zstandard.train_dictionary,
                16384,
                samples
            )
            await asyncio.to_thread(zdict.precompute_compress, level=3)
            return (zdict, 3)
        ```
        """

    @property
    def compression_training_high_watermark(self) -> int:
        """After we reach the low watermark and coordinate a compression dictionary, we retain
        those samples and wait until we reach this watermark to train the dictionary again.

        The low watermark gets us to some level of compression reasonably quickly, and the high
        watermark gets us to a more accurate dictionary once there's enough data to work with.

        A reasonable value is 10mb
        """

    async def train_compression_dict_high_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        """Trains a compression dictionary using the samples whose combined size is at least the
        `compression_training_high_watermark` size and tells us what compression level to use.

        Typically, this is something like

        ```python
        import zstandard
        import asyncio

        async def train_compression_dict_low_watermark(samples: List[bytes]) -> zstandard.ZstdCompressionDict:
            zdict = await asyncio.to_thread(
                zstandard.train_dictionary,
                65536,
                samples
            )
            await asyncio.to_thread(zdict.precompute_compress, level=10)
            return (zdict, 10)
        ```
        """

    @property
    def compression_retrain_interval_seconds(self) -> int:
        """How long in seconds between rebuilding a compression dictionary for very long-lived
        websocket connections.

        Especially when there are timestamps within the message body, the compression dictionary
        needs occasional refreshing to remain effective.

        A reasonable value is 1 day
        """

    @property
    def decompression_max_window_size(self) -> int:
        """
        Sets an upper limit on the window size for decompression operations
        in kibibytes. This setting can be used to prevent large memory
        allocations for inputs using large compression windows.

        Use 0 for no limit.

        A reasonable value is 0 for no limit. Alternatively, it should be 8mb if
        trying to match the zstandard minimum decoder requirements. The
        remaining alternative would be as high as the server can bear, noting
        that this much memory may be allocated by every websocket connection up
        to 3 times (once for standard decompression without a custom dictionary,
        once for the most recent custom dictionary, and once for the second most
        recent custom dictionary). If disabling training, a websocket will use
        up to 1x this memory on decompression buffers.

        WARN:
            This should not be considered a security measure. Authorization
            is already passed prior to decompression, and if that is not enough
            to eliminate adversarial payloads, then disable compression.
        """


class CompressionConfigFromParts:
    """Convenience class that allows you to create a CompressionConfig protocol
    satisfying object from values, using default implementations for the methods
    """

    def __init__(
        self,
        compression_allowed: bool,
        compression_dictionary_by_id: "Dict[int, Tuple[zstandard.ZstdCompressionDict, int]]",
        outgoing_max_ws_message_size: Optional[int],
        allow_training: bool,
        compression_min_size: int,
        compression_trained_max_size: int,
        compression_training_low_watermark: int,
        compression_training_high_watermark: int,
        compression_retrain_interval_seconds: int,
        decompression_max_window_size: int,
    ):
        if compression_allowed:
            try:
                importlib.import_module("zstandard")
            except ImportError:
                raise ValueError(
                    "Compression is allowed, but zstandard is not available. "
                    "Set compression_allowed=False to disable compression, or "
                    "`pip install zstandard` to enable it."
                )

        if 0 in compression_dictionary_by_id:
            raise ValueError("Dictionary ID 0 is reserved for no compression")

        if 1 in compression_dictionary_by_id:
            raise ValueError(
                "Dictionary ID 1 is reserved for not using a compression dictionary"
            )

        self.compression_allowed = compression_allowed
        self.compression_dictionary_by_id = compression_dictionary_by_id
        self.outgoing_max_ws_message_size = outgoing_max_ws_message_size
        self.allow_training = allow_training
        self.compression_min_size = compression_min_size
        self.compression_trained_max_size = compression_trained_max_size
        self.compression_training_low_watermark = compression_training_low_watermark
        self.compression_training_high_watermark = compression_training_high_watermark
        self.compression_retrain_interval_seconds = compression_retrain_interval_seconds
        self.decompression_max_window_size = decompression_max_window_size

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /
    ) -> "Optional[Tuple[zstandard.ZstdCompressionDict, int]]":
        return self.compression_dictionary_by_id.get(dictionary_id)

    async def train_compression_dict_low_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        zdict = await asyncio.to_thread(zstandard.train_dictionary, 16384, samples)
        await asyncio.to_thread(zdict.precompute_compress, level=3)
        return (zdict, 3)

    async def train_compression_dict_high_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        zdict = await asyncio.to_thread(zstandard.train_dictionary, 65536, samples)
        await asyncio.to_thread(zdict.precompute_compress, level=10)
        return (zdict, 10)


class Config(
    AuthConfig, DBConfig, GenericConfig, MissedRetryConfig, CompressionConfig, Protocol
):
    """The injected behavior required for the lonelypss to operate. This is
    generally generated for you using one of the templates, see the readme for details
    """


class ConfigFromParts:
    """Convenience class that combines the three parts of the config into a single object."""

    def __init__(
        self,
        auth: AuthConfig,
        db: DBConfig,
        generic: GenericConfig,
        missed: MissedRetryConfig,
        compression: CompressionConfig,
    ):
        self.auth = auth
        self.db = db
        self.generic = generic
        self.missed = missed
        self.compression = compression

    async def setup_to_broadcaster_auth(self) -> None:
        await self.auth.setup_to_broadcaster_auth()

    async def teardown_to_broadcaster_auth(self) -> None:
        await self.auth.teardown_to_broadcaster_auth()

    async def setup_to_subscriber_auth(self) -> None:
        await self.auth.setup_to_subscriber_auth()

    async def teardown_to_subscriber_auth(self) -> None:
        await self.auth.teardown_to_subscriber_auth()

    async def setup_db(self) -> None:
        await self.db.setup_db()

    async def teardown_db(self) -> None:
        await self.db.teardown_db()

    async def authorize_subscribe_exact(
        self, /, *, url: str, recovery: Optional[str], exact: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_exact(
            url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_subscribe_exact_allowed(
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_subscribe_glob(
        self, /, *, url: str, recovery: Optional[str], glob: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_glob(
            url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_subscribe_glob_allowed(
            url=url, recovery=recovery, glob=glob, now=now, authorization=authorization
        )

    async def authorize_notify(
        self, /, *, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_notify(
            topic=topic, message_sha512=message_sha512, now=now
        )

    async def is_notify_allowed(
        self,
        /,
        *,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_notify_allowed(
            topic=topic,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_configure(
        self,
        /,
        *,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_configure(
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
        )

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_stateful_configure_allowed(message=message, now=now)

    async def authorize_check_subscriptions(
        self, /, *, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_check_subscriptions(url=url, now=now)

    async def is_check_subscriptions_allowed(
        self, /, *, url: str, now: float, authorization: Optional[str]
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_check_subscriptions_allowed(
            url=url, now=now, authorization=authorization
        )

    async def authorize_set_subscriptions(
        self, /, *, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_set_subscriptions(
            url=url, strong_etag=strong_etag, now=now
        )

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_set_subscriptions_allowed(
            url=url,
            strong_etag=strong_etag,
            subscriptions=subscriptions,
            now=now,
            authorization=authorization,
        )

    async def authorize_receive(
        self, /, *, url: str, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_receive(
            url=url, topic=topic, message_sha512=message_sha512, now=now
        )

    async def is_receive_allowed(
        self,
        /,
        *,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_receive_allowed(
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def authorize_missed(
        self, /, *, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_missed(recovery=recovery, topic=topic, now=now)

    async def is_missed_allowed(
        self,
        /,
        *,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_missed_allowed(
            recovery=recovery, topic=topic, now=now, authorization=authorization
        )

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_confirm_configure(
            broadcaster_nonce=broadcaster_nonce, now=now
        )

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.auth.is_stateful_confirm_configure_allowed(
            message=message, now=now
        )

    async def subscribe_exact(
        self, /, *, url: str, recovery: Optional[str], exact: bytes
    ) -> Literal["success", "conflict", "unavailable"]:
        return await self.db.subscribe_exact(url=url, recovery=recovery, exact=exact)

    async def unsubscribe_exact(
        self, /, *, url: str, exact: bytes
    ) -> Literal["success", "not_found", "unavailable"]:
        return await self.db.unsubscribe_exact(url=url, exact=exact)

    async def subscribe_glob(
        self, /, *, url: str, recovery: Optional[str], glob: str
    ) -> Literal["success", "conflict", "unavailable"]:
        return await self.db.subscribe_glob(url=url, recovery=recovery, glob=glob)

    async def unsubscribe_glob(
        self, /, *, url: str, glob: str
    ) -> Literal["success", "not_found", "unavailable"]:
        return await self.db.unsubscribe_glob(url=url, glob=glob)

    def get_subscribers(self, /, *, topic: bytes) -> AsyncIterable[SubscriberInfo]:
        return self.db.get_subscribers(topic=topic)

    async def check_subscriptions(self, /, *, url: str) -> StrongEtag:
        return await self.db.check_subscriptions(url=url)

    async def set_subscriptions(
        self,
        /,
        *,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
    ) -> Literal["success", "unavailable"]:
        return await self.db.set_subscriptions(
            url=url, strong_etag=strong_etag, subscriptions=subscriptions
        )

    async def upsert_missed(
        self, /, *, info: MissedInfo
    ) -> Literal["success", "unavailable"]:
        return await self.db.upsert_missed(info=info)

    def get_overdue_missed_with_lock(
        self, /, *, now: float
    ) -> AsyncIterable[LockedMissedInfo]:
        return self.db.get_overdue_missed_with_lock(now=now)

    @property
    def message_body_spool_size(self) -> int:
        return self.generic.message_body_spool_size

    @property
    def outgoing_http_timeout_total(self) -> Optional[float]:
        return self.generic.outgoing_http_timeout_total

    @property
    def outgoing_http_timeout_connect(self) -> Optional[float]:
        return self.generic.outgoing_http_timeout_connect

    @property
    def outgoing_http_timeout_sock_read(self) -> Optional[float]:
        return self.generic.outgoing_http_timeout_sock_read

    @property
    def outgoing_http_timeout_sock_connect(self) -> Optional[float]:
        return self.generic.outgoing_http_timeout_sock_connect

    @property
    def websocket_accept_timeout(self) -> Optional[float]:
        return self.generic.websocket_accept_timeout

    @property
    def websocket_max_pending_sends(self) -> Optional[int]:
        return self.generic.websocket_max_pending_sends

    @property
    def websocket_max_unprocessed_receives(self) -> Optional[int]:
        return self.generic.websocket_max_unprocessed_receives

    @property
    def websocket_large_direct_send_timeout(self) -> Optional[float]:
        return self.generic.websocket_large_direct_send_timeout

    @property
    def websocket_send_max_unacknowledged(self) -> Optional[int]:
        return self.generic.websocket_send_max_unacknowledged

    async def get_delay_for_next_missed_retry(
        self, /, *, receive_url: str, missed_url: str, topic: bytes, attempts: int
    ) -> Optional[float]:
        return await self.missed.get_delay_for_next_missed_retry(
            receive_url=receive_url,
            missed_url=missed_url,
            topic=topic,
            attempts=attempts,
        )

    @property
    def compression_allowed(self) -> bool:
        return self.compression.compression_allowed

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /
    ) -> "Optional[Tuple[zstandard.ZstdCompressionDict, int]]":
        return await self.compression.get_compression_dictionary_by_id(dictionary_id)

    @property
    def outgoing_max_ws_message_size(self) -> Optional[int]:
        return self.compression.outgoing_max_ws_message_size

    @property
    def allow_training(self) -> bool:
        return self.compression.allow_training

    @property
    def compression_min_size(self) -> int:
        return self.compression.compression_min_size

    @property
    def compression_trained_max_size(self) -> int:
        return self.compression.compression_trained_max_size

    @property
    def compression_training_low_watermark(self) -> int:
        return self.compression.compression_training_low_watermark

    async def train_compression_dict_low_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        return await self.compression.train_compression_dict_low_watermark(samples)

    @property
    def compression_training_high_watermark(self) -> int:
        return self.compression.compression_training_high_watermark

    async def train_compression_dict_high_watermark(
        self, /, samples: List[bytes]
    ) -> "Tuple[zstandard.ZstdCompressionDict, int]":
        return await self.compression.train_compression_dict_high_watermark(samples)

    @property
    def compression_retrain_interval_seconds(self) -> int:
        return self.compression.compression_retrain_interval_seconds

    @property
    def decompression_max_window_size(self) -> int:
        return self.compression.decompression_max_window_size

    @property
    def websocket_minimal_headers(self) -> bool:
        return self.generic.websocket_minimal_headers

    @property
    def sweep_missed_interval(self) -> float:
        return self.generic.sweep_missed_interval


if TYPE_CHECKING:
    _a: Type[GenericConfig] = GenericConfigFromValues
    _b: Type[CompressionConfig] = CompressionConfigFromParts
    _c: Type[MissedRetryConfig] = MissedRetryStandard
    _d: Type[Config] = ConfigFromParts



import base64
import hashlib
import io
import json
import logging
import tempfile
import time
from dataclasses import dataclass
from enum import Enum, auto
from typing import Annotated, Dict, Literal, Optional, Union

import aiohttp
from fastapi import APIRouter, Header, Request, Response
from pydantic import BaseModel, Field

from lonelypss.config.config import (
    Config,
    MissedInfo,
    SubscriberInfo,
    SubscriberInfoType,
)
from lonelypss.middleware.config import get_config_from_request
from lonelypss.util.close_guarded_io import CloseGuardedIO
from lonelypss.util.sync_io import SyncIOBaseLikeIO, read_exact


class NotifyResponse(BaseModel):
    notified: int = Field(description="The number of subscribers successfully notified")


router = APIRouter()


@router.post(
    "/v1/notify",
    status_code=200,
    response_model=NotifyResponse,
    responses={
        "400": {"description": "The body was not formatted correctly"},
        "401": {"description": "Authorization header is required but not provided"},
        "403": {"description": "Authorization header is provided but invalid"},
        "500": {"description": "Unexpected error occurred"},
        "503": {"description": "Service is unavailable, try again soon"},
    },
)
async def notify(
    request: Request, authorization: Annotated[Optional[str], Header()] = None
) -> Response:
    """Sends the given message to subscribers for the given topic. The body should be
    formatted as the following sequence:

    - 2 bytes: the length of the topic, big-endian, unsigned
    - N bytes: the topic
    - 64 bytes: the sha-512 hash of the message. will be rechecked
    - 8 bytes: the length of the message, big-endian, unsigned
    - M bytes: the message to send. must have the same hash as the provided hash

    The response has one of the following status codes, where the body is arbitrary
    unless otherwise specified.

    - 200 Okay: subscribers were notified. Response body is in JSON format,
      containing the `notified` key with the number of subscribers notified.
    - 400 Bad Request: the body was not formatted correctly
    - 401 Unauthorized: authorization is required but not provided
    - 403 Forbidden: authorization is provided but invalid
    - 500 Internal Server Error: unexpected error occurred
    - 503 Service Unavailable: servce (generally, database) is unavailable
    """
    config = get_config_from_request(request)

    with tempfile.SpooledTemporaryFile(
        max_size=config.message_body_spool_size, mode="w+b"
    ) as request_body:
        read_length = 0
        saw_end = False

        stream_iter = request.stream().__aiter__()
        while True:
            try:
                chunk = await stream_iter.__anext__()
            except StopAsyncIteration:
                saw_end = True
                break

            request_body.write(chunk)
            read_length += len(chunk)
            if read_length >= 2 + 65535 + 64 + 8:
                break

        request_body.seek(0)
        topic_length = int.from_bytes(read_exact(request_body, 2), "big")
        topic = read_exact(request_body, topic_length)
        message_hash = read_exact(request_body, 64)
        message_length = int.from_bytes(read_exact(request_body, 8), "big")

        auth_at = time.time()
        auth_result = await config.is_notify_allowed(
            topic=topic,
            message_sha512=message_hash,
            now=auth_at,
            authorization=authorization,
        )

        if auth_result == "unauthorized":
            return Response(status_code=401)
        elif auth_result == "forbidden":
            return Response(status_code=403)
        elif auth_result == "unavailable":
            return Response(status_code=503)
        elif auth_result != "ok":
            return Response(status_code=500)

        hasher = hashlib.sha512()

        while True:
            chunk = request_body.read(io.DEFAULT_BUFFER_SIZE)
            if not chunk:
                break
            hasher.update(chunk)

        if not saw_end:
            while True:
                try:
                    chunk = await stream_iter.__anext__()
                except StopAsyncIteration:
                    saw_end = True
                    break

                hasher.update(chunk)
                request_body.write(chunk)
                read_length += len(chunk)

                if read_length > 2 + topic_length + 64 + 8 + message_length:
                    return Response(status_code=400)

        if read_length != 2 + topic_length + 64 + 8 + message_length:
            return Response(status_code=400)

        actual_hash = hasher.digest()
        if actual_hash != message_hash:
            return Response(status_code=400)

        request_body.seek(2 + topic_length + 64 + 8)
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=config.outgoing_http_timeout_total,
                connect=config.outgoing_http_timeout_connect,
                sock_read=config.outgoing_http_timeout_sock_read,
                sock_connect=config.outgoing_http_timeout_sock_connect,
            )
        ) as session:
            notify_result = await handle_trusted_notify(
                topic,
                request_body,
                config=config,
                session=session,
                content_length=message_length,
                sha512=actual_hash,
            )

        if notify_result.type == TrustedNotifyResultType.UNAVAILABLE:
            return Response(status_code=503)

        return Response(
            status_code=200,
            content=NotifyResponse.__pydantic_serializer__.to_json(
                NotifyResponse(notified=notify_result.succeeded)
            ),
            headers={
                "Content-Type": "application/json; charset=utf-8",
            },
        )


class TrustedNotifyResultType(Enum):
    UNAVAILABLE = auto()
    """We had trouble accessing the data store and may not have attempted all
    subscribers
    """
    OK = auto()
    """We at least attempted all subscribers"""


@dataclass
class TrustedNotifyResultOK:
    type: Literal[TrustedNotifyResultType.OK]
    """discriminator type"""
    succeeded: int
    """The number of subscribers we reached"""
    failed: int
    """The number of subscribers we could not reach"""


@dataclass
class TrustedNotifyResultUnavailable:
    type: Literal[TrustedNotifyResultType.UNAVAILABLE]
    """discriminator type"""
    partial_succeeded: int
    """The number of subscribers we reached"""
    partial_failed: int
    """The number of subscribers we could not reach"""


TrustedNotifyResult = Union[TrustedNotifyResultOK, TrustedNotifyResultUnavailable]


async def _handle_missed(
    config: Config,
    topic: bytes,
    subscriber: SubscriberInfo,
) -> None:
    if subscriber.type == SubscriberInfoType.UNAVAILABLE:
        return
    if subscriber.recovery is None:
        return

    next_retry_at = await config.get_delay_for_next_missed_retry(
        receive_url=subscriber.url,
        missed_url=subscriber.recovery,
        topic=topic,
        attempts=0,
    )
    if next_retry_at is not None:
        await config.upsert_missed(
            info=MissedInfo(
                topic=topic,
                attempts=0,
                next_retry_at=next_retry_at,
                subscriber_info=subscriber,
            )
        )


async def handle_trusted_notify(
    topic: bytes,
    data: SyncIOBaseLikeIO,
    /,
    *,
    config: Config,
    session: aiohttp.ClientSession,
    content_length: int,
    sha512: bytes,
) -> TrustedNotifyResult:
    """Notifies subscribers to the given topic with the given data.

    Args:
        topic (bytes): the topic the message was sent to
        data (file-like, readable, bytes): the message that was sent
        config (Config): the broadcaster configuration to use
        session (aiohttp.ClientSession): the aiohttp client session to
            send requests to clients in
        content_length (int): the length of the message in bytes. The stream
            MUST not have more than this amount of data in it, as we will read
            past this length if it does
        sha512 (bytes): the sha512 hash of the content (64 bytes)
    """
    succeeded = 0
    failed = 0
    headers: Dict[str, str] = {
        "Content-Type": "application/octet-stream",
        "Content-Length": str(content_length),
        "Repr-Digest": f"sha-512={base64.b64encode(sha512).decode('ascii')}",
        "X-Topic": base64.b64encode(topic).decode("ascii"),
    }

    message_starts_at = data.tell()
    guarded_request_body = CloseGuardedIO(data)

    async for subscriber in config.get_subscribers(topic=topic):
        if subscriber.type == SubscriberInfoType.UNAVAILABLE:
            return TrustedNotifyResultUnavailable(
                type=TrustedNotifyResultType.UNAVAILABLE,
                partial_succeeded=succeeded,
                partial_failed=failed,
            )

        failed += 1
        my_authorization = await config.authorize_receive(
            url=subscriber.url, topic=topic, message_sha512=sha512, now=time.time()
        )
        if my_authorization is None:
            headers.pop("Authorization", None)
        else:
            headers["Authorization"] = my_authorization

        data.seek(message_starts_at)
        handled_missed = False
        try:
            async with session.post(
                subscriber.url,
                data=guarded_request_body,
                headers=headers,
            ) as resp:
                if resp.ok:
                    logging.debug(
                        f"Successfully notified {subscriber.url} about {topic!r}"
                    )

                    num_subscribers = 1

                    content_type = resp.headers.get("Content-Type")
                    if content_type is not None and content_type.startswith(
                        "application/json"
                    ):
                        content = await resp.json()
                        if (
                            isinstance(content, dict)
                            and isinstance(content.get("subscribers"), int)
                            and content["subscribers"] >= 0
                        ):
                            num_subscribers = content["subscribers"]

                    succeeded += num_subscribers
                    failed -= 1

                else:
                    logging.warning(
                        f"Failed to notify {subscriber.url} about {topic!r}: {resp.status}"
                    )
                    handled_missed = True
                    await _handle_missed(config, topic, subscriber)

                    if resp.status >= 400 and resp.status < 500:
                        content_type = resp.headers.get("Content-Type")
                        if content_type is not None and content_type.startswith(
                            "application/json"
                        ):
                            content = await resp.json()
                            if (
                                isinstance(content, dict)
                                and content.get("unsubscribe") is True
                            ):
                                logging.info(
                                    f"Unsubscribing {subscriber.url} from {topic!r} due to response: {json.dumps(content)}"
                                )

                                if subscriber.type == SubscriberInfoType.EXACT:
                                    await config.unsubscribe_exact(
                                        url=subscriber.url, exact=topic
                                    )
                                else:
                                    await config.unsubscribe_glob(
                                        url=subscriber.url, glob=subscriber.glob
                                    )

        except aiohttp.ClientError:
            logging.error(
                f"Failed to notify {subscriber.url} about {topic!r}", exc_info=True
            )
            if not handled_missed:
                await _handle_missed(config, topic, subscriber)

    return TrustedNotifyResultOK(
        type=TrustedNotifyResultType.OK,
        succeeded=succeeded,
        failed=failed,
    )



import math
import secrets
import sqlite3
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from lonelypsp.stateless.make_strong_etag import (
    GlobAndRecovery,
    StrongEtag,
    TopicAndRecovery,
    create_strong_etag_generator,
)
from lonelypsp.util.bounded_deque import BoundedDeque

from lonelypss.config.config import (
    LockedMissedInfo,
    MissedInfo,
    MutableMissedInfo,
    SubscriberInfo,
    SubscriberInfoExact,
    SubscriberInfoGlob,
    SubscriberInfoType,
)
from lonelypss.config.set_subscriptions_info import SetSubscriptionsInfo
from lonelypss.ws.handlers.open.errors import combine_multiple_exceptions

if TYPE_CHECKING:
    from lonelypss.config.config import DBConfig

import sys

if sys.version_info < (3, 10):
    from enum import Enum, auto
    from typing import Any, TypeVar, overload

    class _NotSet(Enum):
        NOT_SET = auto()

    T = TypeVar("T")
    D = TypeVar("D")

    @overload
    async def anext(iterable: AsyncIterator[T], /) -> T: ...

    @overload
    async def anext(iterable: AsyncIterator[T], default: D, /) -> Union[T, D]: ...

    async def anext(
        iterable: AsyncIterator[T],
        default: Any = _NotSet.NOT_SET,
        /,
    ) -> Any:
        try:
            return await iterable.__anext__()
        except StopAsyncIteration:
            if default is _NotSet.NOT_SET:
                raise
            return default


class SqliteLockedMissedInfo:
    """Implementation of LockedMissedInfo for Sqlite"""

    def __init__(
        self, parent: "SqliteDBConfig", info: MissedInfo, lock_id: int
    ) -> None:
        self.parent = parent
        self.info = info
        self.lock_id = lock_id
        self.released = False

    async def release(
        self, /, *, new_info: Optional[MutableMissedInfo] = None
    ) -> Literal["conflict", "unavailable", "ok"]:
        if self.released:
            return "conflict"

        assert self.parent.cursor is not None, "db not setup"

        if self.info.subscriber_info.type == SubscriberInfoType.EXACT:
            if new_info is None:
                return self._release_exact_with_delete(
                    self.parent.cursor,
                    self.info.subscriber_info,
                )
            return self._release_exact_with_update(
                self.parent.cursor,
                self.info.subscriber_info,
                new_info,
            )

        if new_info is None:
            return self._release_glob_with_delete(
                self.parent.cursor,
                self.info.subscriber_info,
            )

        return self._release_glob_with_update(
            self.parent.cursor,
            self.info.subscriber_info,
            new_info,
        )

    def _release_exact_with_delete(
        self,
        cursor: sqlite3.Cursor,
        subscriber_info: SubscriberInfoExact,
    ) -> Literal["conflict", "unavailable", "ok"]:
        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        cursor.execute(
            "DELETE FROM httppubsub_subscription_exacts "
            "WHERE url=? AND exact=? AND missed_lock_id=?",
            (subscriber_info.url, self.info.topic, self.lock_id),
        )
        deleted = cursor.rowcount > 0
        cursor.execute("COMMIT")

        self.released = True
        return "ok" if deleted else "conflict"

    def _release_exact_with_update(
        self,
        cursor: sqlite3.Cursor,
        subscriber_info: SubscriberInfoExact,
        new_info: MutableMissedInfo,
    ) -> Literal["conflict", "unavailable", "ok"]:
        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        cursor.execute(
            "UPDATE httppubsub_subscription_exacts "
            "SET missed_attempts=?, missed_next_retry_at=?, missed_lock_id=NULL, missed_lock_expires_at=NULL "
            "WHERE url=? AND exact=? AND missed_lock_id=?",
            (
                new_info.attempts,
                math.ceil(new_info.next_retry_at),
                subscriber_info.url,
                self.info.topic,
                self.lock_id,
            ),
        )
        updated = cursor.rowcount > 0
        cursor.execute("COMMIT")

        self.released = True
        return "ok" if updated else "conflict"

    def _release_glob_with_delete(
        self,
        cursor: sqlite3.Cursor,
        subscriber_info: SubscriberInfoGlob,
    ) -> Literal["conflict", "unavailable", "ok"]:
        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        cursor.execute(
            "DELETE FROM httppubsub_subscription_globs "
            "WHERE url=? AND glob=? AND missed_lock_id=?",
            (subscriber_info.url, subscriber_info.glob, self.lock_id),
        )
        deleted = cursor.rowcount > 0
        cursor.execute("COMMIT")

        self.released = True
        return "ok" if deleted else "conflict"

    def _release_glob_with_update(
        self,
        cursor: sqlite3.Cursor,
        subscriber_info: SubscriberInfoGlob,
        new_info: MutableMissedInfo,
    ) -> Literal["conflict", "unavailable", "ok"]:
        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        cursor.execute(
            "UPDATE httppubsub_subscription_globs "
            "SET missed_attempts=?, missed_next_retry_at=?, missed_lock_id=NULL, missed_lock_expires_at=NULL "
            "WHERE url=? AND glob=? AND missed_lock_id=?",
            (
                new_info.attempts,
                math.ceil(new_info.next_retry_at),
                subscriber_info.url,
                subscriber_info.glob,
                self.lock_id,
            ),
        )
        updated = cursor.rowcount > 0
        cursor.execute("COMMIT")

        self.released = True
        return "ok" if updated else "conflict"


class SqliteDBConfig:
    """Implements the DBConfig protocol using a locally hosted SQLite database."""

    def __init__(self, database: str, missed_lock_time: int = 30) -> None:
        self.database = database
        """The database url. You can pass `:memory:` to create a SQLite database that 
        exists only in memory, otherwise, this is typically the path to a sqlite file
        (usually has the `db` extension).
        """

        self.missed_lock_time = missed_lock_time
        """The number of seconds a missed lock should be held for before it is
        considered expired and can be stolen by other processes or coroutines,
        which could lead to duplicate MISSED messages but will generally not
        be harmful
        """

        self.connection: Optional[sqlite3.Connection] = None
        """If we've been setup, the current connection to the database."""

        self.cursor: Optional[sqlite3.Cursor] = None
        """If we've been setup, the current cursor to the database."""

    async def setup_db(self) -> None:
        assert self.connection is None, "sqlite db config is not re-entrant"
        connection = sqlite3.connect(self.database, isolation_level=None)
        cursor = connection.cursor()

        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
            cursor.execute(
                """
    CREATE TABLE IF NOT EXISTS httppubsub_subscription_exacts (
        id INTEGER PRIMARY KEY,
        url TEXT NOT NULL,
        exact BLOB NOT NULL,
        missed_url TEXT NULL,
        missed_lock_id INTEGER NULL,
        missed_lock_expires_at INTEGER NULL,
        missed_attempts INTEGER NULL,
        missed_next_retry_at INTEGER NULL
    )
                """
            )
            cursor.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_httppubsub_subscription_exacts_exact_url ON httppubsub_subscription_exacts (exact, url)"
            )
            # get_overdue_missed_with_lock
            cursor.execute(
                """
    CREATE INDEX IF NOT EXISTS idx_httppubsub_subscription_exacts_next_retry_at 
    ON httppubsub_subscription_exacts (missed_next_retry_at) 
    WHERE missed_next_retry_at IS NOT NULL
                """
            )
            # delete_missed
            cursor.execute(
                """
    CREATE INDEX IF NOT EXISTS idx_httppubsub_subscription_exacts_missed_url_topic
    ON httppubsub_subscription_exacts (missed_url, exact) WHERE missed_url IS NOT NULL
                """
            )

            cursor.execute(
                """
    CREATE TABLE IF NOT EXISTS httppubsub_subscription_globs (
        id INTEGER PRIMARY KEY,
        url TEXT NOT NULL,
        glob TEXT NOT NULL,
        missed_url TEXT NULL,
        missed_lock_id INTEGER NULL,
        missed_lock_expires_at INTEGER NULL,
        missed_topic BLOB NULL,
        missed_attempts INTEGER NULL,
        missed_next_retry_at INTEGER NULL
    )
                """
            )
            cursor.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_httppubsub_subscription_globs_glob_url ON httppubsub_subscription_globs (glob, url)"
            )
            # get_overdue_missed_with_lock
            cursor.execute(
                """
    CREATE INDEX IF NOT EXISTS idx_httppubsub_subscription_globs_next_retry_at
    ON httppubsub_subscription_globs (missed_next_retry_at)
    WHERE missed_next_retry_at IS NOT NULL
                """
            )
            # delete_missed
            cursor.execute(
                """
    CREATE INDEX IF NOT EXISTS idx_httppubsub_subscription_globs_missed_url_topic
    ON httppubsub_subscription_globs (missed_url, missed_topic) WHERE missed_url IS NOT NULL
                """
            )
            cursor.execute("COMMIT")
        except BaseException as og_cause:
            related: List[BaseException] = []
            try:
                cursor.close()
            except BaseException as e:
                related.append(e)

            try:
                connection.close()
            except BaseException as e:
                related.append(e)

            if related:
                raise combine_multiple_exceptions(
                    "failed to initialize sqlite db, then failed to close",
                    [og_cause, *related],
                )
            raise

        self.connection = connection
        self.cursor = cursor

    async def teardown_db(self) -> None:
        exc1: Optional[BaseException] = None
        exc2: Optional[BaseException] = None
        if self.cursor is not None:
            try:
                self.cursor.close()
            except BaseException as e:
                exc1 = e
            self.cursor = None

        if self.connection is not None:
            try:
                self.connection.close()
            except BaseException as e:
                exc2 = e
            self.connection = None

        if exc2 is not None:
            raise exc2
        if exc1 is not None:
            raise exc1

    async def subscribe_exact(
        self, /, *, url: str, recovery: Optional[str], exact: bytes
    ) -> Literal["success", "conflict", "unavailable"]:
        assert self.cursor is not None, "db not setup"
        try:
            self.cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        self.cursor.execute(
            """
INSERT INTO httppubsub_subscription_exacts (
    url, 
    exact, 
    missed_url,
    missed_lock_id,
    missed_lock_expires_at,
    missed_attempts, 
    missed_next_retry_at
)
SELECT ?, ?, ?, NULL, NULL, NULL, NULL
WHERE
    NOT EXISTS (
        SELECT 1 FROM httppubsub_subscription_exacts AS hse
        WHERE hse.url = ? AND hse.exact = ?
    )
            """,
            (url, exact, recovery, url, exact),
        )
        inserted = self.cursor.rowcount > 0
        self.cursor.execute("COMMIT")

        return "success" if inserted else "conflict"

    async def unsubscribe_exact(
        self, /, *, url: str, exact: bytes
    ) -> Literal["success", "not_found", "unavailable"]:
        assert self.cursor is not None, "db not setup"

        try:
            self.cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        self.cursor.execute(
            "DELETE FROM httppubsub_subscription_exacts WHERE url = ? AND exact = ?",
            (url, exact),
        )
        deleted = self.cursor.rowcount > 0
        self.cursor.execute("COMMIT")

        return "success" if deleted else "not_found"

    async def subscribe_glob(
        self, /, *, url: str, recovery: Optional[str], glob: str
    ) -> Literal["success", "conflict", "unavailable"]:
        assert self.cursor is not None, "db not setup"
        try:
            self.cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        self.cursor.execute(
            """
INSERT INTO httppubsub_subscription_globs (
    url,
    glob,
    missed_url,
    missed_lock_id,
    missed_lock_expires_at,
    missed_topic,
    missed_attempts,
    missed_next_retry_at
)
SELECT ?, ?, ?, NULL, NULL, NULL, NULL, NULL
WHERE
    NOT EXISTS (
        SELECT 1 FROM httppubsub_subscription_globs AS hsg
        WHERE hsg.url = ? AND hsg.glob = ?
    )
            """,
            (url, glob, recovery, url, glob),
        )
        inserted = self.cursor.rowcount > 0
        self.cursor.execute("COMMIT")

        return "success" if inserted else "conflict"

    async def unsubscribe_glob(
        self, /, *, url: str, glob: str
    ) -> Literal["success", "not_found", "unavailable"]:
        assert self.cursor is not None, "db not setup"
        try:
            self.cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"

        self.cursor.execute(
            "DELETE FROM httppubsub_subscription_globs WHERE url = ? AND glob = ?",
            (url, glob),
        )
        deleted = self.cursor.rowcount > 0
        self.cursor.execute("COMMIT")
        return "success" if deleted else "not_found"

    async def get_subscribers(
        self, /, *, topic: bytes
    ) -> AsyncIterable[SubscriberInfo]:
        assert self.connection is not None, "db not setup"
        cursor = self.connection.cursor()
        try:
            cursor.execute("BEGIN DEFERRED TRANSACTION")
            cursor.execute(
                "SELECT url, missed_url FROM httppubsub_subscription_exacts WHERE exact = ?",
                (topic,),
            )
            topic_rows = cast(List[Tuple[str, Optional[str]]], cursor.fetchall())
            cursor.execute("COMMIT")

            for url, recovery in topic_rows:
                yield SubscriberInfoExact(
                    type=SubscriberInfoType.EXACT, url=url, recovery=recovery
                )

            cursor.execute("BEGIN DEFERRED TRANSACTION")
            cursor.execute(
                "SELECT glob, url, missed_url FROM httppubsub_subscription_globs WHERE ? GLOB glob",
                (topic,),
            )
            glob_rows = cast(List[Tuple[str, str, Optional[str]]], cursor.fetchall())
            cursor.execute("COMMIT")
            for glob, url, recovery in glob_rows:
                yield SubscriberInfoGlob(
                    type=SubscriberInfoType.GLOB, glob=glob, url=url, recovery=recovery
                )
        finally:
            cursor.close()

    async def check_subscriptions(self, /, *, url: str) -> StrongEtag:
        assert self.connection is not None, "db not setup"

        topics_gen = create_strong_etag_generator(url)

        cursor = self.connection.cursor()
        try:
            last_topic: Optional[bytes] = None
            while True:
                cursor.execute("BEGIN DEFERRED TRANSACTION")
                cursor.execute(
                    "SELECT exact, missed_url "
                    "FROM httppubsub_subscription_exacts "
                    "WHERE url = ?"
                    " AND (? IS NULL OR exact > ?) "
                    "ORDER BY exact ASC "
                    "LIMIT 100",
                    (url, last_topic, last_topic),
                )

                topics_batch = cast(
                    List[Tuple[bytes, Optional[str]]], cursor.fetchall()
                )
                cursor.execute("COMMIT")
                if not topics_batch:
                    break

                topics_gen.add_topic(
                    *(TopicAndRecovery(*topic) for topic in topics_batch)
                )

                last_topic = topics_batch[-1][0]

            globs_gen = topics_gen.finish_topics()
            del topics_gen

            last_glob: Optional[str] = None
            while True:
                cursor.execute("BEGIN DEFERRED TRANSACTION")
                cursor.execute(
                    "SELECT glob, missed_url "
                    "FROM httppubsub_subscription_globs "
                    "WHERE url = ?"
                    " AND (? IS NULL OR glob > ?) "
                    "ORDER BY glob ASC "
                    "LIMIT 100",
                    (url, last_glob, last_glob),
                )

                globs_batch = cast(List[Tuple[str, Optional[str]]], cursor.fetchall())
                cursor.execute("COMMIT")
                if not globs_batch:
                    break

                globs_gen.add_glob(*(GlobAndRecovery(*glob) for glob in globs_batch))

                last_glob = globs_batch[-1][0]

            return globs_gen.finish()
        finally:
            cursor.close()

    async def set_subscriptions(
        self,
        /,
        *,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
    ) -> Literal["success", "unavailable"]:
        assert self.connection is not None, "db not setup"

        cursor = self.connection.cursor()
        try:
            await self._set_topics(
                cursor=cursor, url=url, desired_topics_iter=subscriptions.topics()
            )
            await self._set_globs(
                cursor=cursor, url=url, desired_globs_iter=subscriptions.globs()
            )
            return "success"
        finally:
            cursor.close()

    async def _set_topics(
        self,
        /,
        *,
        cursor: sqlite3.Cursor,
        url: str,
        desired_topics_iter: AsyncIterator[TopicAndRecovery],
    ) -> None:
        actual_topics_iter = _ExistingTopicsIter(cursor, url)

        next_desired: Optional[TopicAndRecovery] = await anext(
            desired_topics_iter, None
        )
        next_actual: Optional[bytes] = next(actual_topics_iter, None)

        while next_desired is not None or next_actual is not None:
            if next_desired is None:
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute(
                    "DELETE FROM httppubsub_subscription_exacts "
                    "WHERE url = ? AND exact >= ?",
                    (url, next_actual),
                )
                cursor.execute("COMMIT")
                next_actual = None
            elif next_actual is None or next_desired.topic < next_actual:
                await self.subscribe_exact(
                    url=url, recovery=next_desired.recovery, exact=next_desired.topic
                )
                next_desired = await anext(desired_topics_iter, None)
            elif next_desired.topic > next_actual:
                await self.unsubscribe_exact(url=url, exact=next_actual)
                next_actual = next(actual_topics_iter, None)
            else:
                next_desired = await anext(desired_topics_iter, None)
                next_actual = next(actual_topics_iter, None)

    async def _set_globs(
        self,
        /,
        *,
        cursor: sqlite3.Cursor,
        url: str,
        desired_globs_iter: AsyncIterator[GlobAndRecovery],
    ) -> None:
        actual_globs_iter = _ExistingGlobsIter(cursor, url)

        next_desired: Optional[GlobAndRecovery] = await anext(desired_globs_iter, None)
        next_actual: Optional[str] = next(actual_globs_iter, None)

        while next_desired is not None or next_actual is not None:
            if next_desired is None:
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute(
                    "DELETE FROM httppubsub_subscription_globs "
                    "WHERE url = ? AND glob >= ?",
                    (url, next_actual),
                )
                cursor.execute("COMMIT")
                next_actual = None
            elif next_actual is None or next_desired.glob < next_actual:
                await self.subscribe_glob(
                    url=url, recovery=next_desired.recovery, glob=next_desired.glob
                )
                next_desired = await anext(desired_globs_iter, None)
            elif next_desired.glob > next_actual:
                await self.unsubscribe_glob(url=url, glob=next_actual)
                next_actual = next(actual_globs_iter, None)
            else:
                next_desired = await anext(desired_globs_iter, None)
                next_actual = next(actual_globs_iter, None)

    async def upsert_missed(
        self, /, *, info: MissedInfo
    ) -> Literal["success", "unavailable"]:
        assert self.cursor is not None and self.connection is not None, "db not setup"

        cursor = self.cursor

        if info.subscriber_info.type == SubscriberInfoType.EXACT:
            try:
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
            except sqlite3.Error:
                return "unavailable"

            cursor.execute(
                """
UPDATE httppubsub_subscription_exacts
SET
    missed_url = ?,
    missed_attempts = ?,
    missed_next_retry_at = ?,
    missed_lock_id = NULL,
    missed_lock_expires_at = NULL
WHERE
    url = ? AND exact = ?
                """,
                (
                    info.subscriber_info.recovery,
                    info.attempts,
                    math.ceil(info.next_retry_at),
                    info.subscriber_info.url,
                    info.topic,
                ),
            )
            cursor.execute("COMMIT")
            return "success"

        try:
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        except sqlite3.Error:
            return "unavailable"
        cursor.execute(
            """
UPDATE httppubsub_subscription_globs
SET
    missed_url = ?,
    missed_topic = ?,
    missed_attempts = ?,
    missed_next_retry_at = ?,
    missed_lock_id = NULL,
    missed_lock_expires_at = NULL
WHERE
    url = ? AND glob = ?
            """,
            (
                info.subscriber_info.recovery,
                info.topic,
                info.attempts,
                math.ceil(info.next_retry_at),
                info.subscriber_info.url,
                info.subscriber_info.glob,
            ),
        )
        cursor.execute("COMMIT")
        return "success"

    def _make_missed_id(self) -> int:
        return int.from_bytes(secrets.token_bytes(8), "big", signed=True)

    async def get_overdue_missed_with_lock(
        self, /, *, now: float
    ) -> AsyncIterable[LockedMissedInfo]:
        assert self.connection is not None, "db not setup"

        cursor = self.connection.cursor()
        try:
            last_next_retry_at: Optional[float] = None
            while True:
                # most of the time this will need a write lock, and it's more obvious
                # what the possible implications are if we error SQLITE_BUSY here rather
                # than later
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute(
                    """
SELECT id, url, exact, missed_url, missed_attempts, missed_next_retry_at
FROM httppubsub_subscription_exacts
WHERE
    missed_next_retry_at IS NOT NULL
    AND missed_next_retry_at <= ?
    AND (? IS NULL OR missed_next_retry_at > ?)
    AND (missed_lock_expires_at IS NULL OR missed_lock_expires_at < ?)
ORDER BY missed_next_retry_at ASC
LIMIT 1
                    """,
                    (now, last_next_retry_at, last_next_retry_at, now),
                )
                exact_row = cast(
                    Optional[Tuple[int, str, bytes, str, int, int]], cursor.fetchone()
                )
                if exact_row is None:
                    cursor.execute("COMMIT")
                    break

                row_id, url, exact, recovery, attempts, next_retry_at = exact_row
                lock_id = self._make_missed_id()
                cursor.execute(
                    "UPDATE httppubsub_subscription_exacts "
                    "SET"
                    " missed_lock_id = ?,"
                    " missed_lock_expires_at = ? "
                    "WHERE id = ?",
                    (lock_id, now + self.missed_lock_time, row_id),
                )
                if cursor.rowcount != 1:
                    cursor.execute("ROLLBACK")
                    raise RuntimeError("failed to lock row despite held transaction")
                cursor.execute("COMMIT")
                yield SqliteLockedMissedInfo(
                    parent=self,
                    info=MissedInfo(
                        topic=exact,
                        attempts=attempts,
                        next_retry_at=next_retry_at,
                        subscriber_info=SubscriberInfoExact(
                            type=SubscriberInfoType.EXACT, url=url, recovery=recovery
                        ),
                    ),
                    lock_id=lock_id,
                )

            last_next_retry_at = None
            while True:
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute(
                    """
SELECT id, url, glob, missed_url, missed_topic, missed_attempts, missed_next_retry_at
FROM httppubsub_subscription_globs
WHERE
    missed_next_retry_at IS NOT NULL
    AND missed_next_retry_at <= ?
    AND (? IS NULL OR missed_next_retry_at > ?)
    AND (missed_lock_expires_at IS NULL OR missed_lock_expires_at < ?)
ORDER BY missed_next_retry_at ASC
LIMIT 1
                    """,
                    (now, last_next_retry_at, last_next_retry_at, now),
                )
                glob_row = cast(
                    Optional[Tuple[int, str, str, str, bytes, int, int]],
                    cursor.fetchone(),
                )
                if glob_row is None:
                    cursor.execute("COMMIT")
                    break

                row_id, url, glob, recovery, topic, attempts, next_retry_at = glob_row
                lock_id = self._make_missed_id()
                cursor.execute(
                    "UPDATE httppubsub_subscription_globs "
                    "SET"
                    " missed_lock_id = ?,"
                    " missed_lock_expires_at = ? "
                    "WHERE id = ?",
                    (lock_id, now + self.missed_lock_time, row_id),
                )
                if cursor.rowcount != 1:
                    cursor.execute("ROLLBACK")
                    raise RuntimeError("failed to lock row despite held transaction")
                cursor.execute("COMMIT")
                yield SqliteLockedMissedInfo(
                    parent=self,
                    info=MissedInfo(
                        topic=topic,
                        attempts=attempts,
                        next_retry_at=next_retry_at,
                        subscriber_info=SubscriberInfoGlob(
                            type=SubscriberInfoType.GLOB,
                            glob=glob,
                            url=url,
                            recovery=recovery,
                        ),
                    ),
                    lock_id=lock_id,
                )

        finally:
            cursor.close()


class _ExistingTopicsIter:
    def __init__(self, cursor: sqlite3.Cursor, url: str, batch_size: int = 16) -> None:
        self.cursor = cursor
        self.url = url
        self.batch_size = batch_size

        self.batch_remaining: BoundedDeque[bytes] = BoundedDeque(maxlen=batch_size)
        self.last_topic: Optional[bytes] = None
        self.at_last_batch: bool = False

    def __iter__(self) -> "_ExistingTopicsIter":
        return self

    def __next__(self) -> bytes:
        if not self.batch_remaining and not self.at_last_batch:
            self._get_next_batch()

        if not self.batch_remaining:
            raise StopIteration

        return self.batch_remaining.popleft()

    def _get_next_batch(self) -> None:
        self.cursor.execute(
            "SELECT exact "
            "FROM httppubsub_subscription_exacts "
            "WHERE url = ?"
            " AND (? IS NULL OR exact > ?) "
            "ORDER BY exact ASC "
            "LIMIT ?",
            (self.url, self.last_topic, self.last_topic, self.batch_size),
        )
        raw_batch: List[Tuple[bytes]] = self.cursor.fetchall()
        self.batch_remaining.ensure_space_for(len(raw_batch))
        for row in raw_batch:
            self.batch_remaining.append(row[0])
        self.at_last_batch = len(raw_batch) < self.batch_size


class _ExistingGlobsIter:
    def __init__(self, cursor: sqlite3.Cursor, url: str, batch_size: int = 16) -> None:
        self.cursor = cursor
        self.url = url
        self.batch_size = batch_size

        self.batch_remaining: BoundedDeque[str] = BoundedDeque(maxlen=batch_size)
        self.last_glob: Optional[str] = None
        self.at_last_batch: bool = False

    def __iter__(self) -> "_ExistingGlobsIter":
        return self

    def __next__(self) -> str:
        if not self.batch_remaining and not self.at_last_batch:
            self._get_next_batch()

        if not self.batch_remaining:
            raise StopIteration

        return self.batch_remaining.popleft()

    def _get_next_batch(self) -> None:
        self.cursor.execute(
            "SELECT glob "
            "FROM httppubsub_subscription_globs "
            "WHERE url = ?"
            " AND (? IS NULL OR glob > ?) "
            "ORDER BY glob ASC "
            "LIMIT ?",
            (self.url, self.last_glob, self.last_glob, self.batch_size),
        )
        raw_batch: List[Tuple[str]] = self.cursor.fetchall()
        self.batch_remaining.ensure_space_for(len(raw_batch))
        for row in raw_batch:
            self.batch_remaining.append(row[0])
        self.at_last_batch = len(raw_batch) < self.batch_size


if TYPE_CHECKING:
    _a: Type[LockedMissedInfo] = SqliteLockedMissedInfo
    _b: Type[DBConfig] = SqliteDBConfig
    _c: Type[Iterable[bytes]] = _ExistingTopicsIter
    _d: Type[Iterable[str]] = _ExistingGlobsIter
